unit MainGUI;

///////////////////////////////////////////////////////////////////////
/////   The GUI used by the user, and control of the programme    /////
/////                                                             /////
/////    Functions:                                               /////
/////    - View of fitness records in grid format                 /////
/////    - Entry of new records                                   /////
/////    - Edit of existing records                               /////
/////    - Filter and sort the view                               /////
/////    - Graph the fitness records                              /////
/////    - Export of data                                         /////
/////    - Print fitness records                                  /////
/////    - Add/edit the data tables                               /////
/////    - configure user preferences                             /////
///////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////
/////    Units:                                                   /////
/////    - vsqlite3 (with type Tvsqlite3)                         /////
/////      - access to sqlite3 database                           /////
/////      - written so that a change of database driver entails  /////
/////        a change of this unit. The rest of the programme     /////
/////        remains unchanges.                                   /////
/////     - vtabledata (Tvtabledata)                              /////
/////       - manipulation of the metadata                        /////
/////     - vFilter (Tvfilter)                                    /////
/////       - controls the filtering of the view of data          /////
/////     - vAddEdit (Taddedit)                                   /////
/////       - controls the adding or editing of records           /////
/////     - vGraph (Tvgraph)                                      /////
/////       - controls the display of the Graph                   /////
/////     - vTypes                                                /////
/////       - Types, variables and constants common to the units  /////
/////     - vCourses                                              /////
/////       - Manipulation of the Courses data                    /////
/////     - vINI                                                  /////
/////       - read/write from the INI files (vfitness, vfilter)   /////
///////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////
/////     Notes:                                                  /////
/////     - GUI is designed for MS Windows                        /////
/////     - GUI is modified for other operating systems           /////
/////       programatically                                       /////
/////     - Second data field is always plotted on X axis         /////
///////////////////////////////////////////////////////////////////////

{$mode objfpc}{$H+}

{ TODO : Check that there are no dependencies (eg if deleting a course, make sure that there are none of those courses in the rides table) }
{ TODO : Restore from backups. Choose from a list of backups showing date of backup, and the last change. }
{ TODO : Write an 'all tables' list showing dates and type of exercise (with an INI setting selecting which tables to include) }
{ TODO : Put in error-checking into LoadTableNames in case format of vfitness.ini is incorrect }
{ TODO : Write 'new' routine for new table/field in editTablesGUI }
{ TODO : in Options, allow to save state (of graph, of sorting on stringgrid, of filter in use) }
{ TODO : when adding a new record, the chkcourses is ticked. }

interface

uses
  Classes, Forms, Controls, Dialogs, StdCtrls, Menus, LCLtype, EditBtn,
  ComCtrls, Grids, ExtCtrls, TAGraph, TASeries, vsqlite3, vINI,
  vTableData, vStringData, vCourses, vFilter, AddEditGUI,  vTypes, SysUtils,
  vGraphs, vUtils, EditTablesGUI;

type

  { TfrmMainGUI }

  TfrmMainGUI = class(TForm)
    btnExit: TButton;
    btnAdd: TButton;
    btnEdit: TButton;
    btnDelete: TButton;
    btnLast: TButton;
    btnNext: TButton;
    btnPrevious: TButton;
    btnFirst: TButton;
    btnStats: TButton;
    btnMap: TButton;
    btnEditDataStructure: TButton;
    cmbTableNames: TComboBox;
    gbxFilter: TGroupBox;
    grpFilter: TGroupBox;
    lblDatabases: TLabel;
    {$IFDEF DARWIN}
      AppMenu     : TMenuItem;
      AppAboutCmd : TMenuItem;
      AppSep1Cmd  : TMenuItem;
      AppPrefCmd  : TMenuItem;
    {$ENDIF}
    MainMenu: TMainMenu;
    MenuItem1: TMenuItem;
    MenuItem2: TMenuItem;
    MenuItem3: TMenuItem;
    mnuHelpAbout: TMenuItem;
    mnuHelp: TMenuItem;
    mnuGoLast: TMenuItem;
    mnuGoNext: TMenuItem;
    mnuGoPrevious: TMenuItem;
    mnuGoFirst: TMenuItem;
    mnuGo: TMenuItem;
    mnuView: TMenuItem;
    mnuEditOptions: TMenuItem;
    mnuEditSelectAll: TMenuItem;
    mnuEditPaste: TMenuItem;
    mnuEditCopy: TMenuItem;
    mnuEditCut: TMenuItem;
    mnuEditUndo: TMenuItem;
    mnuEdit: TMenuItem;
    mnuFIleExit: TMenuItem;
    mnuFIlePrint: TMenuItem;
    mnuFileSave: TMenuItem;
    mnuFIle: TMenuItem;
    PageControl1: TPageControl;
    SgdResults: TStringGrid;
    sgdStats: TStringGrid;
    tabFilter: TTabSheet;
    tabGraphs: TTabSheet;
    tabOptions: TTabSheet;
    tabPartners: TTabSheet;
    tabResults: TTabSheet;
    procedure btnAddClick(Sender: TObject);
    procedure btnDeleteClick(Sender: TObject);
    procedure btnEditClick(Sender: TObject);
    procedure btnEditDataStructureClick(Sender: TObject);
    procedure btnExitClick(Sender: TObject);
    procedure btnFirstClick(Sender: TObject);
    procedure btnLastClick(Sender: TObject);
    procedure btnNextClick(Sender: TObject);
    procedure btnPreviousClick(Sender: TObject);
    procedure btnStatsClick(Sender: TObject);
    // Change table after user selection.
    procedure cmbTableNamesSelect(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure FormClose(Sender: TObject; var CloseAction: TCloseAction);
    // Initialization of the programme
    procedure FormCreate(Sender: TObject);
    // Display the 'about box'
    procedure mnuAboutClick(Sender: TObject);
    procedure mnuFileSaveClick(Sender: TObject);
    // Display and allow changes to 'Preferences' or  'Options'
    procedure OptionsCmdClick(Sender: TObject);
    // When changing the size of the Grid, force the bottom row to display at
    // the bottom of the grid.
    procedure ForceBottowRowToGridBottom;
    procedure PageControl1Change(Sender: TObject);
  private
    FDataChanged: boolean;
    procedure SetDataChanged(const AValue: boolean);
  private
    { private declarations }
    // Variables for the classes declared in other units.
    vvStringData: TvStringData;
    vvSqlite3: TvSqlite3;
    vvINI: TvINI;
    vvCourses: TvCourses;
    vvTableData: TvTableData;
    vvFilter: TvFilter;
    vvGraphs: TvGraphs;
    frmAddEditGUI: TfrmAddEditGUI;
    // Date format...changed to 'dd/mm/yy', but changed back to default when programme ends
    DateFormat: string;
    DateFormatSeparator: char;
    // The name of the Bundle/Application.
    BundleName : String;
    // The path including filename to the database
    DBpath: string;
    // The path to the Support files (eg INI files, Database)
    DataPath: string;
    // The current row of sgdResults
    CurrRow: integer;
    // Modify the GUI regardless of OS platform
    procedure ModifyGUI;
    // Modify the GUI to have an 'Apple' look and feel to it.
    procedure ModifyGUI4Apple;
    // Make the height of the GUI longer for Apple
    procedure ModifyFormHeight;
    // Modify the botton row of buttons for Apple
    procedure ModifyBottomButtons;
    // Modify the Menus for Apple
    procedure ModifyMenus;
    // Assign Shortcut Keys to Menus
    procedure AssignShortMenuShortcuts;
    // Open my classes that this form is dependant upon
    procedure OpenDependantUnits;
    // Close my classes
    procedure CloseDependantUnits;
    // Get the path to the database
    function GetDBpath():string;
    // Get the path to the support files, such as the INI files and the DB
    function GetDataPath():string;
    // Initialize variables
    procedure InitVars;
    // Load All the Controls with data
    procedure LoadControls;
    // Load TableNames Combo box with table names
    procedure LoadTableNames;
    // Prepare for Adding or Editing an entry
    procedure AddEditEntry(const NewRecord:boolean);
    // Return the 'Fields' from the current row of the string grid, if editing an existing record
    function GetFields(NewRecord: boolean): vFieldArrayType;
    property DataChanged: boolean read FDataChanged write SetDataChanged;
  public
    { public declarations }
  end; 

var
  frmMainGUI: TfrmMainGUI;

implementation

{$R *.lfm}

{ TfrmMainGUI }

uses AboutBox;

procedure TfrmMainGUI.FormCreate(Sender: TObject);
begin
  InitVars;
  ModifyGUI;
  OpenDependantUnits;
  LoadControls;
end;

// Only show the AboutBox modally if on a Windows platform.
procedure TfrmMainGUI.mnuAboutClick(Sender: TObject);
begin
  {$IFDEF DARWIN}
    if Assigned(frmAboutBox) then Exit;
    frmAboutBox := TfrmAboutBox.Create(Application);
    frmAboutBox.Show;
  {$ELSE}
    frmAboutBox := TfrmAboutBox.Create(Application);
    try
      frmAboutBox.ShowModal;
    finally
      frmAboutBox.Free;
    end;
  {$ENDIF}
end;

procedure TfrmMainGUI.mnuFileSaveClick(Sender: TObject);
begin
  ShowMessage('Saved!');
end;

procedure TfrmMainGUI.OptionsCmdClick(Sender: TObject);
begin
  PageControl1.ActivePage := tabOptions;
end;

procedure TfrmMainGUI.ForceBottowRowToGridBottom;
begin
  sgdResults.Row := 1;
  sgdResults.Row := CurrRow;
end;

procedure TfrmMainGUI.PageControl1Change(Sender: TObject);
var
  Display: boolean;
begin
  Display             := PageControl1.ActivePage = tabResults;
  btnFirst.Visible    := Display;
  btnPrevious.Visible := Display;
  btnNext.Visible     := Display;
  btnLast.Visible     := Display;
  btnDelete.Visible   := Display;
  btnAdd.Visible      := Display;
  btnEdit.Visible     := Display;
  btnMap.Visible      := Display;
  btnStats.Visible    := Display;
end;

procedure TfrmMainGUI.SetDataChanged(const AValue: boolean);
begin
  FDataChanged:=AValue;
end;

procedure TfrmMainGUI.btnExitClick(Sender: TObject);
begin
  Close;
end;

procedure TfrmMainGUI.btnEditClick(Sender: TObject);
begin
  AddEditEntry(False);
end;

procedure TfrmMainGUI.btnEditDataStructureClick(Sender: TObject);
begin
  CloseDependantUnits;
  frmEditTablesGUI := TfrmEditTablesGUI.Create(DBpath, DataPath);
  self.Hide;
  try
    frmEditTablesGUI.ShowModal;
  finally
    frmEditTablesGUI.Destroy;
    OpenDependantUnits;
    LoadControls;
  end;
  self.Show;
  frmEditTablesGUI := nil;
end;

procedure TfrmMainGUI.btnAddClick(Sender: TObject);
begin
  AddEditEntry(True);
end;

procedure TfrmMainGUI.btnDeleteClick(Sender: TObject);
begin
    if Application.MessageBox(PChar('Are you sure you want to delete the data for ' +
      SgdResults.Cells[1, SgdResults.Row] + '?'), 'Warning -- Deletion pending',
      MB_ICONQUESTION + MB_YESNO + MB_DEFBUTTON3) = idYes then
        vvStringData.DeleteData;
end;

procedure TfrmMainGUI.btnFirstClick(Sender: TObject);
begin
  vvStringData.GoFirst;
end;

procedure TfrmMainGUI.btnLastClick(Sender: TObject);
begin
  vvStringData.GoLast;
end;

procedure TfrmMainGUI.btnNextClick(Sender: TObject);
begin
  vvStringData.GoNext;
end;

procedure TfrmMainGUI.btnPreviousClick(Sender: TObject);
begin
  vvStringData.GoPrevious;
end;

procedure TfrmMainGUI.btnStatsClick(Sender: TObject);
begin
  // Bug in String Grid on Mac, when changing size of Grid.
  // CurrRow is used again in ForceBottowRowToGridBottom
  CurrRow := sgdResults.row;
  if btnStats.Caption = 'Stats' then
  begin
    sgdResults.Height := tabResults.Height - sgdResults.Top - sgdStats.Height;
    btnStats.Caption := 'List';
  end
  else
  begin
    sgdResults.Height := tabResults.Height - sgdResults.Top - 5;
    btnStats.Caption := 'Stats';
  end;
  sgdStats.Visible := not sgdStats.Visible;
  ForceBottowRowToGridBottom;
end;

// Load the TableNames combobox with User-friendly names, with associated real names.
procedure TfrmMainGUI.cmbTableNamesSelect(Sender: TObject);
var
  Table: string;
begin
  Table := vvTableData.GetTableName(cmbTableNames.Items[cmbTableNames.ItemIndex]);
  vvFilter.LoadTable(Table);
  if vvTableData.GetGraphed() then
    tabGraphs.TabVisible := True
  else
    tabGraphs.TabVisible := False;
end;

procedure TfrmMainGUI.FormActivate(Sender: TObject);
begin
  vvFilter.ActivateFilter;
end;

procedure TfrmMainGUI.FormClose(Sender: TObject; var CloseAction: TCloseAction);
begin
  DefaultFormatSettings.DateSeparator   := DateFormatSeparator;
  DefaultFormatSettings.ShortDateFormat := DateFormat;
  CloseDependantUnits;
  if DataChanged then
    BackupDB(DataPath);
  Unused(CloseAction);
end;

procedure TfrmMainGUI.ModifyGUI;
begin
  ModifyGUI4Apple;
end;

procedure TfrmMainGUI.ModifyGUI4Apple;
begin
  ModifyFormHeight;
  ModifyBottomButtons;
  ModifyMenus;
end;

procedure TfrmMainGUI.ModifyFormHeight;
begin
  {$ifdef Darwin}
  frmMainGui.Height:=535;
  PageControl1.Height:=382;
  {$endif}
end;

procedure TfrmMainGUI.ModifyBottomButtons;
var
  i, group1, group2, group3, increment:integer;
begin
  {
  There are 4 groups of buttons:
  - group1: First, Previous, Next, Last
  - group2: Delete, Edit, Add
  - group3: Map, Stats
  - group4: Exit

  'group1' is the leftmost start of Group 1
  'group2' is the leftmost start of Group 2
  'group3' is the leftmost start of Group 3
  'increment' is the distance to the next button in the group
  }
  group1 := 16;
  group2 := 290;
  group3 := 560;
  increment := 64;
  for i := 0 to frmMainGUI.ControlCount -1 do begin
    if frmMainGUI.Controls[i] is TButton then
      with frmMainGUI.Controls[i] do begin
          // Change the 'Height' to 22, and Increase the Width to accommodate the Apple Font
          Height:=22;
          Width:=62;
          Top := frmMainGui.Height - 58;
          // Reposition the controls because of the increased width
          case Name of
            'btnFirst'    : Left := group1;
            'btnPrevious' : Left := group1 + increment;
            'btnNext'     : Left := group1 + increment*2;
            'btnLast'     : Left := group1 + increment*3;
            'btnDelete'   : Left := group2;
            'btnEdit'     : Left := group2 + increment;
            'btnAdd'      : Left := group2 + increment*2;
            'btnMap'      : Left := group3;
            'btnStats'    : Left := group3 + increment;
            'btnExit'     : Left := 760;
          end;
        end;
  end;
end;

procedure TfrmMainGUI.ModifyMenus;
begin
  {$ifdef Darwin}
    // Add 'Apple' Menu
    AppMenu := TMenuItem.Create(Self);  // Application menu
    AppMenu.Caption := #$EF#$A3#$BF;    // Unicode Apple logo char
    MainMenu.Items.Insert(0, AppMenu);

    // Add 'About' as item in application menu
    AppAboutCmd := TMenuItem.Create(Self);
    AppAboutCmd.Caption := 'About ' + BundleName;
    AppAboutCmd.OnClick := @mnuAboutClick;
    AppMenu.Add(AppAboutCmd);

    // Add Separator to application menu
    AppSep1Cmd := TMenuItem.Create(Self);
    AppSep1Cmd.Caption := '-';
    AppMenu.Add(AppSep1Cmd);

    // Add 'Preferences'/'Options' to application menu
    AppPrefCmd          := TMenuItem.Create(Self);
    AppPrefCmd.Caption  := 'Preferences...';
    AppPrefCmd.Shortcut := ShortCut(VK_OEM_COMMA, [ssMeta]);
    AppPrefCmd.OnClick  := @OptionsCmdClick;
    AppMenu.Add(AppPrefCmd);

    // Adjust height of main form because the menu is not part of the form
    frmMainGui.Height := 507;

    // Remove 'Exit' from 'File' menu, and 'About' from 'Help' menu, and 'Options' from 'Edit' menu.
    mnuFileExit.Visible    := False;
    mnuHelpAbout.Visible   := False;
    mnuEditOptions.Visible :=False;

    AssignShortMenuShortcuts;
  {$endif}
end;

procedure TfrmMainGUI.AssignShortMenuShortcuts;
begin
  mnuEditCopy.Shortcut      := ShortCut(VK_C, [ssMeta]);    // Command-C
  mnuEditPaste.Shortcut     := ShortCut(VK_V, [ssMeta]);    // Command-V
  mnuEditUndo.Shortcut      := ShortCut(VK_Z, [ssMeta]);    // Command-Z
  mnuEditCut.Shortcut       := ShortCut(VK_X, [ssMeta]);    // Command-X
  mnuEditSelectAll.Shortcut := ShortCut(VK_A, [ssMeta]);    // Command-A
  //mnuGoPrevious.Shortcut    := ShortCut(VK_UP, []);         // Up
  //mnuGoNext.Shortcut        := ShortCut(VK_DOWN, []);       // Down
  mnuFIle.Shortcut      := ShortCut(VK_S, [ssMeta]);    // Command-S
  mnuFilePrint.Shortcut     := ShortCut(VK_P, [ssCtrl]);    // Ctrl-P
end;

procedure TfrmMainGUI.OpenDependantUnits;
begin
  vvsqlite3    := Tvsqlite3.Create(GetDBpath);
  vvINI        := TvINI.Create(DataPath);
  vvCourses    := TvCourses.Create(vvSqlite3);
  vvTableData  := TvTableData.Create(vvINI, vvSqlite3);
  vvStringData := TvStringData.Create(sgdResults, sgdStats, btnMap, vvsqlite3,vvTableData, vvCourses);
  vvGraphs     := TvGraphs.Create(tabGraphs, vvStringData, vvTableData);
  vvFilter     := TvFilter.Create(grpFilter,tabFilter, gbxFilter, vvCourses, vvStringData, vvTableData,
    vvINI, vvGraphs);
end;

procedure TfrmMainGUI.CloseDependantUnits;
begin
  vvGraphs.Destroy;
  vvFilter.Destroy;
  vvStringData.Destroy;
  vvTableData.Destroy;
  vvCourses.Destroy;
  vvINI.Destroy;
  vvsqlite3.Destroy;
end;

function TfrmMainGUI.GetDBpath():string;
begin
  DBpath := DataPath + '/vfitnessDB';
  Result := DBpath;
end;

// This is pretty rudimentary.
function TfrmMainGUI.GetDataPath(): string;
var
  temp:string;
begin
  {$ifdef Darwin}
    temp := '~/Library/vfitness';
  {$else}
    temp := GetAppConfigDir(False);
  {$endif}
  // 'ExpandFileName' expands the '~' to the user's home directory.
  Result := ExpandFileName(temp);
end;

procedure TfrmMainGUI.InitVars;
begin
  BundleName := 'vfitness';
  DataPath   := GetDataPath;
  DateFormat := DefaultFormatSettings.ShortDateFormat;
  DateFormatSeparator := DefaultFormatSettings.DateSeparator;
  DefaultFormatSettings.DateSeparator := '/';
  DefaultFormatSettings.ShortDateFormat := 'DD/MM/YYYY';
  btnStatsClick(nil);
  DataChanged := False;
end;

procedure TfrmMainGUI.LoadControls;
begin
  LoadTableNames;
  {$ifdef windows}
  cmbTableNamesSelect(nil);
  {$endif}
  PageControl1.ActivePage:=TabResults;
end;

procedure TfrmMainGUI.LoadTableNames;
begin
  vvTableData.LoadUFtableNames(cmbTableNames);
end;

procedure TfrmMainGUI.AddEditEntry(const NewRecord:boolean);
begin
  frmAddEditGUI := TfrmAddEditGUI.Create(Application, vvTableData, vvCourses, vvStringData, GetFields(NewRecord));
  self.Hide;
  try
    frmAddEditGUI.ShowModal;
  finally
    if not DataChanged then
      DataChanged := frmAddEditGUI.DataChanged;
    frmAddEditGUI.Destroy;
  end;
  self.Show;
  frmAddEditGUI := nil;
end;

function TfrmMainGUI.GetFields(NewRecord: boolean): vFieldArrayType;
var
  Fields: vFieldArrayType;
begin
  if NewRecord then
    Fields := vvTableData.GetDefaultValues
  else
    Fields := vvStringData.OneRecord;
  Result := Fields;
end;

end.

