unit vCourses;

{$mode objfpc}{$H+}

////////////////////////////////////////////////////////////////////////
////       Unit providing operations on the 'Courses' tables        ////
////////////////////////////////////////////////////////////////////////

interface

uses
  Classes, SysUtils, StdCtrls, vTypes, vSqlite3;

type
  { TvCourses }

  TvCourses = class
  private
    FNumberOfCourses: integer;
    Courses: VCourseArrayType;
    vvSqlite3: TvSqlite3;
    CourseComboBoxes: Tlist;
    function GetNumberOfFields: integer;
    procedure SetNumberOfCourses(const AValue: integer);
    property NumberOfCourses: integer read FNumberOfCourses write SetNumberOfCourses;
    // Store the Course Names in a combobox
    procedure LoadCourseNames(cmb: TComboBox);
    // Load the Courses Table from the Database to memory
    procedure LoadCourseTable;
  public
    constructor Create(Sqlite3:TvSqlite3);
    destructor Destroy; override;
    // Return the name of the Course, given the Course ID
    function CourseName(const CID: integer): string;
    // Return the ID of the Course, given the Course Name
    function CourseID(const Cname: string): string;
    // Register a comboBox so TvCourses can update it when data changed
    procedure RegisterComboBox(var cmb:TComboBox);
    procedure unRegisterComboBox(var cmb:TComboBox);
    // Reload all of the registered ComboBoxes with the changed Courses Table
    procedure Reload;
    // Return the actual name of the Courses table.
    function TableName: string;
    // Return the number of fields in Courses
    property NumberOfFields: integer read GetNumberOfFields;
    // Return the Course Description given the course name
    function CourseDescription(const Cname: string): string;
    // Return the Course Distance given the course name
    function CourseDistance(const Cname: string): string;
    // Return the Course URL given the course name
    function CourseURL(const Cname: string): string;
    // Return True if the Course Cname has a URL
    function  URLExists(const Cname: string): boolean;
  end;

implementation

uses StrUtils;

{ TvCourses }

function TvCourses.GetNumberOfFields: integer;
begin
  Result := 4;
end;

procedure TvCourses.SetNumberOfCourses(const AValue: integer);
begin
  FNumberOfCourses:=AValue;
end;

constructor TvCourses.Create(Sqlite3: TvSqlite3);
begin
  vvSqlite3 := Sqlite3;
  CourseComboBoxes := Tlist.Create;
  LoadCourseTable;
end;

destructor TvCourses.Destroy;
begin
  CourseComboBoxes.Free;
  inherited Destroy;
end;

function TvCourses.CourseName(const CID: integer): string;
var
  i: integer = 0;
begin
  while i < NumberOfCourses do
  begin
    if StrToInt(Courses[i].ID) = CID then
      break;
    i := i + 1;
  end;
  Result := ifthen(i < NumberOfCourses, Courses[i].Name, '');
end;

function TvCourses.CourseID(const Cname: string): string;
var
  i: integer = 0;
begin
  while i < NumberOfCourses do
  begin
    if CompareText(Cname,Courses[i].Name) = 0 then
      break;
    i := i + 1;
  end;
  Result := ifthen(i < NumberOfCourses, Courses[i].ID, '');
end;

procedure TvCourses.RegisterComboBox(var cmb: TComboBox);
begin
  CourseComboBoxes.Add(cmb);
  LoadCourseNames(cmb);
end;

procedure TvCourses.unRegisterComboBox(var cmb: TComboBox);
begin
  CourseComboBoxes.Delete(CourseComboBoxes.IndexOf(cmb));
end;

procedure TvCourses.Reload;
var
  i: integer;
begin
  LoadCourseTable;
  for i := 0 to CourseComboBoxes.Count - 1 do
    LoadCourseNames(TcomboBox(CourseComboBoxes[i]));
end;

procedure TvCourses.LoadCourseNames(cmb: TComboBox);
var
  i, j: integer;
begin
  j := cmb.ItemIndex;
  cmb.Clear;
  for i := 0 to NumberOfCourses -1 do
    cmb.Items.Add(Courses[i].Name);
  cmb.ItemIndex:=j;
end;

procedure TvCourses.LoadCourseTable;
begin
  NumberOfCourses := vvSqlite3.LoadCourseTable(Courses);
end;

function TvCourses.TableName: string;
begin
  Result := 'Courses';
end;

function TvCourses.CourseDescription(const Cname: string): string;
var
  i: integer = 0;
begin
  while i < NumberOfCourses do
  begin
    if CompareText(Cname,Courses[i].Name) = 0 then
      break;
    i := i + 1;
  end;
  Result := ifthen(i < NumberOfCourses, Courses[i].Description, '');
end;

function TvCourses.CourseDistance(const Cname: string): string;
var
  i: integer = 0;
begin
  while i < NumberOfCourses do
  begin
    if CompareText(Cname,Courses[i].Name) = 0 then
      break;
    i := i + 1;
  end;
  Result := ifthen(i < NumberOfCourses, Courses[i].Distance, '');
end;

function TvCourses.CourseURL(const Cname: string): string;
var
  i: integer = 0;
begin
  while i < NumberOfCourses do
  begin
    if CompareText(Cname,Courses[i].Name) = 0 then
      break;
    i := i + 1;
  end;
  Result := ifthen(i < NumberOfCourses, Courses[i].URL, '');
end;

function TvCourses.URLExists(const Cname: string): boolean;
begin
  Result := CourseURL(Cname) <> '';
end;

end.

