unit vFilter;

{$mode objfpc}{$H+}

////////////////////////////////////////////////////////////////////////
////                              vFilter                           ////
////                                                                ////
////  - Class to handle the input of Filters (eg Speed > 25kph)     ////
////                                                                ////
////////////////////////////////////////////////////////////////////////

interface

uses
  Classes, Dialogs, SysUtils, ComCtrls, StdCtrls, EditBtn, Controls, vCourses, vStringData, vTableData,
  vGraphs, vINI, vTypes;

const
  MaxFilterControls = 7;
  FilterINI         = 'Filters.ini';

type

  // Record of Controls to Allow user to filter the table
  vFilterControlType = record
    Enabled:           TcheckBox;
    Field, Comparison: TcomboBox;
    Quantity:          Tedit;
    AndOr:             TcomboBox;
  end;

  { TvFilter }

  TvFilter = class
  private
    private
      FCurrentWhere: string;
      // The Parent for the 'instant filters' (course and dates), and
      // the parent for the table-specific filters
      QuickFilterControl, TabControl, OptionsControl: TwinControl;
      Courses:         TvCourses;
      StringData:      TvStringData;
      TableData:       TvTableData;
      // Pointer to filters.ini
      Fini:            TvINI;
      Graphs:          TvGraphs;
      // Used by the CurrentTable property
      FCurrentTable:   String;
      // Quick Filter controls
      cmbCourses:      TcomboBox;
      chkCourses:      TcheckBox;
      lblCourses:      Tlabel;
      chkDateFrom:     TcheckBox;
      dteDateFrom:     TdateEdit;
      lblDateFrom:     Tlabel;
      chkDateTo:       TcheckBox;
      dteDateTo:       TdateEdit;
      lblDateTo:       Tlabel;
      // Load and Save Filter controls
      gbxFilter:       TgroupBox;
      cmbFilter:       TcomboBox;
      btnUseFilter:    Tbutton;
      btnDeleteFilter: Tbutton;
      btnSaveFilter:   Tbutton;
      edtFilter:       Tedit;
      Label1:          Tlabel;
      btnFilter:       Tbutton;
      // Options controls
      chkOptionCourse: TcheckBox;
      chkOptionDates:  TcheckBox;
      // workaround for bug in tdateedit...see DateFromOnChange and CreateWhere
      TdateTo:       string;
      TdateFrom:     string;
      // Array of Filter Controls
      FilterControls: array [0..MaxFilterControls] of vFilterControlType;
      // Options Record
      Options: TvFilterOptions;
      // Used by the CurrentWhere property
      procedure SetCurrentWhere(const AValue: string);
      // Used by the CurrentTable property
      procedure SetCurrentTable(const AValue: string);
      // Create the controls
      procedure CreateControls;
      procedure CreateQuickFilterControls;
      procedure CreateSaveFiltersControls;
      procedure CreateTableSpecificControls;
      procedure CreateOptionsControls;
      // Create the Where Statement (minus 'where ') and load the table
      procedure CreateWhere;
      procedure AddCourseFilter(var Where: string);
      procedure AddDatesFilter(var Where: string);
      procedure AddTableSpecificFilter(var Where: string);
      // cmbCourses onchange procedure
      procedure CouresOnChange(Sender: Tobject);
      // onchange event procedure for many of the controls
      procedure FilterControlOnChange(Sender: Tobject);
      // onAcceptDate for the TdateEdit controls (used as a replacement for the onchange event
      procedure DateFromOnChange(Sender: Tobject; var Adate: TDateTime; var AcceptDate: boolean);
      procedure DateToOnChange(Sender: Tobject; var Adate: TDateTime; var AcceptDate: boolean);
      property CurrentWhere: string read FCurrentWhere write SetCurrentWhere;
      property CurrentTable:string read FCurrentTable write SetCurrentTable;
      // Display the Course and Date Filter controls, if applicable to the current table
      procedure DisplayQuickFilters;
       // Add another row of Filter Controls
      procedure AddFilterControlRow(Row: integer);
     // Fill the Table-specific Filter Controls with Data
      procedure FillTableSpecificControls(CurrFC: integer);
      // Free Memory used by Table-specific Filter Controls
      procedure DestroyFilterControls;
      // Initialize the Table-specific Controls (eg clear combo boxes, hide all of the rows)
      procedure InitializeTableSpecificControls;
      // onChange event for AndOr combobox
      procedure AndOrOnChange(Sender: TObject);
      // Load the saved Filters into cmbFilter
      procedure LoadFilters;
      // onClick event for btnUseFilter - Use the filter in cmbFilter.
      procedure btnUseFilterClick(Sender: TObject);
      // onClick event for btnSaveFilter - Prepare to save and use the current filter
      procedure btnSaveFilterClick(Sender: TObject);
      // onClick event for btnFilter - Use the filter and show the String Grid.
      procedure btnFilterClick(Sender: TObject);
      // onClick event for btnDelete - Delete the filter in edtFilter.text
      procedure btnDeleteFilterClick(Sender: TObject);
      // Save the current Filter
      function SaveFilter: boolean;
      // Save the Filter Options
      procedure OptionsOnChange(Sender: TObject);
      procedure SaveOptions;
      // Restore the Option selections from the INI file
      procedure RestoreOptions;
public
      constructor Create(QuickCntrl, TabCntrl, OptionsCntrl: TwinControl;Corses: TvCourses;
        StrData:TvStringData; Tdata: TvTableData; FilterINI: TvINI; Grphs: TvGraphs);
      destructor Destroy; override;
      // Work around to prevent the Filter's Controls' onchange event activating until after the cmbTableNames
      // is populated. The procedure is called in the frmMainGUI.formactivate procedure.
      procedure ActivateFilter;
      // The procedure that loads the current table, leading to the Filter controls being
      // loaded, the Graph drawn, and the StringGrid filled with data
      procedure LoadTable(Table: string);
  end;

implementation

{ TvFilter }

uses vUtils, StrUtils;

procedure TvFilter.SetCurrentTable(const AValue: string);
begin
  FCurrentTable:=AValue;
  TableData.TableName:=AValue;
  InitializeTableSpecificControls;
  AddFilterControlRow(0);
  LoadFilters;
end;

procedure TvFilter.SetCurrentWhere(const AValue: string);
begin
  FCurrentWhere:=AValue;
end;

procedure TvFilter.CreateControls;
begin
  CreateQuickFilterControls;
  CreateSaveFiltersControls;
  CreateTableSpecificControls;
  CreateOptionsControls;
end;

procedure TvFilter.CreateQuickFilterControls;
begin
  cmbCourses := TcomboBox.Create(nil);
  with cmbCourses do begin
    Parent  := QuickFilterControl;
    Visible := True;
    Left    := 286;
    Top     := 15;
    Width   := 176;
    Height  := 20;
    Style   := csDropDownList;
    Sorted  := True;
    Courses.RegisterComboBox(cmbCourses);
    ItemIndex := 0;
  end;

  chkCourses := TcheckBox.Create(nil);
  with chkCourses do begin
      Parent  := QuickFilterControl;
      Visible := True;
      Left    := 262;
      Top     := 15;
      Checked := False;
      onChange := @FilterControlOnChange;
  end;

  lblCourses := Tlabel.Create(nil);
  with lblCourses do begin
      Parent  := QuickFilterControl;
      Visible := True;
      Left    := 262;
      Top     := -3;
      Caption := 'Course';
      Autosize:= True;
  end;

  chkDateFrom := TcheckBox.Create(nil);
  with chkDateFrom do begin
      Parent  := QuickFilterControl;
      Visible := True;
      Left    := 526;
      Top     := 15;
      Checked := False;
      OnChange:= @FilterControlOnChange;
  end;

  chkDateTo := TcheckBox.Create(nil);
  with chkDateTo do begin
      Parent  := QuickFilterControl;
      Visible := True;
      Left    := 686;
      Top     := 15;
      Checked := False;
      OnChange:= @FilterControlOnChange;
  end;

  dteDateFrom := TdateEdit.Create(nil);
  with dteDateFrom do begin
      Parent  := QuickFilterControl;
      Visible := True;
      Left    := 550;
      Top     := 15;
      {$ifdef Darwin}
      Width   := 85;
      {$endif}
      DefaultToday := True;
      Text:=TdateFrom;
  end;

  dteDateTo := TdateEdit.Create(nil);
  with dteDateTo do begin
      Parent  := QuickFilterControl;
      Visible := True;
      Left    := 710;
      Top     := 15;
      {$ifdef Darwin}
      Width   := 85;
      {$endif}
      DefaultToday := True;
      Text:=TdateTo;
  end;

  lblDateFrom := Tlabel.Create(nil);
  with lblDateFrom do begin
      Parent  := QuickFilterControl;
      Visible := True;
      Left    := 526;
      Top     := -3;
      Caption := 'Date from';
      Autosize:= True;
  end;

  lblDateTo := Tlabel.Create(nil);
  with lblDateTo do begin
      Parent  := QuickFilterControl;
      Visible := True;
      Left    := 686;
      Top     := -3;
      Caption := 'Date to';
      Autosize:= True;
  end;
end;

procedure TvFilter.CreateSaveFiltersControls;
begin
  gbxFilter := TgroupBox.Create(nil);
  with gbxFilter do begin
    parent   := TabControl;
    Caption  := 'Load or Save filters';
    Height   := 225;
    Hint     := 'Save or load previously used filters';
    Left     := 612;
    ShowHint := True;
    Width    := 185;
    Top      := 16;
  end;

  cmbFilter := TcomboBox.Create(nil);
  with cmbFilter do begin
    parent   := gbxFilter;
    Height   := 20;
    Left     := 14;
    Width    := 152;
    Top      := 16;
    Style    := csDropDownList;
    Sorted   := True;
    ShowHint := True;
  end;

  btnUseFilter := Tbutton.Create(nil);
  with btnUseFilter do begin
    parent   := gbxFilter;
    {$ifdef Darwin}
    Height   := 22;
    Width    := 62;
    {$else}
    Height   := 25;
    Width    := 45;
    {$endif}
    Left     := 14;
    Top      := 48;
    ShowHint := True;
    Caption  := 'Use';
    onClick  := @btnUseFilterClick;
  end;

  btnDeleteFilter := Tbutton.Create(nil);
  with btnDeleteFilter do begin
    parent   := gbxFilter;
    {$ifdef Darwin}
    Height   := 22;
    Width    := 62;
    Left     := 104;
    {$else}
    Height   := 25;
    Width    := 45;
    Left     := 121;
    {$endif}
    Top      := 48;
    ShowHint := True;
    Caption  := 'Delete';
    onClick  := @btnDeleteFilterClick;
  end;

  btnSaveFilter := Tbutton.Create(nil);
  with btnSaveFilter do begin
    parent   := gbxFilter;
    {$ifdef Darwin}
    Height   := 22;
    Width    := 62;
    {$else}
    Height   := 25;
    Width    := 45;
    {$endif}
    Left     := 14;
    Top      := 168;
    ShowHint := True;
    Caption  := 'Save';
    onClick  := @btnSaveFilterClick;
  end;

  edtFilter := Tedit.Create(nil);
  with edtFilter do begin
    parent   := gbxFilter;
    Height   := 23;
    Left     := 14;
    Width    := 152;
    Top      := 136;
    ShowHint := True;
    Autosize := False;
  end;

  Label1 := Tlabel.Create(nil);
  with label1 do begin
    parent   := gbxFilter;
    Height   := 17;
    Left     := 14;
    Width    := 76;
    Top      := 112;
    ShowHint := True;
    Caption  := 'Filter name:';
    Autosize := True;
  end;

  btnFilter := Tbutton.Create(nil);
  with btnFilter do begin
    parent   := TabControl;
    {$ifdef Darwin}
    Height   := 22;
    {$else}
    Height   := 25;
    {$endif}
    Left     := 746;//735;//722;
    Width    := 62; //75;
    Top      := 312;
    Caption  := 'Apply';
    onClick  := @btnFilterClick;
  end;
end;

procedure TvFilter.CreateTableSpecificControls;
var
  i: integer;
begin
  for i := 0 to MaxFilterControls do begin
    FilterControls[i].Enabled := Tcheckbox.Create(nil);
    with FilterControls[i].Enabled do
    begin
      Visible  := False;
      Width    := 20;
      Height   := 19;
      Left     := 20;
      Top      := 24 + 40 * i;
      Parent   := tabControl;
      Checked  := False;
      Hint     := 'Tick the box to include with this filter';
      ShowHint := True;
      TabOrder := 5 * i + 1;
      Tag      := i;
    end;

    FilterControls[i].Field := TcomboBox.Create(nil);
    with FilterControls[i].Field do
    begin
      Visible  := False;
      Width    := 128;
      Height   := 23;
      Left     := FilterControls[i].Enabled.Left + FilterControls[i].Enabled.Width + 28;
      Top      := 24 + 40 * i;
      Parent   := tabControl;
      Style    := csDropDownList;
      Hint     := 'Select the Field to include with this filter';
      ShowHint := True;
      TabOrder := 5 * i + 2;
      Tag      := i;
    end;

    FilterControls[i].Comparison := Tcombobox.Create(nil);
    with FilterControls[i].Comparison do
    begin
      Visible  := False;
      Width    := 70;
      Height   := 23;
      Left     := FilterControls[i].Field.Left + FilterControls[i].Field.Width + 28;
      Top      := 24 + 40 * i;
      Parent   := tabControl;
      Style    := csDropDownList;
      Hint     := 'Select the Comparison Operator to include with this filter';
      ShowHint := True;
      TabOrder := 5 * i + 3;
      Tag      := i;
      Items.Add('<');
      Items.Add('<=');
      Items.Add('=');
      Items.Add('>');
      Items.Add('>=');
      Items.Add('contains');
      Items.Add('excludes');
      Items.Add('not =');
      ItemIndex := 0;
    end;

    FilterControls[i].Quantity := Tedit.Create(nil);
    with FilterControls[i].Quantity do
    begin
      Visible  := False;
      Width    := 104;
      Height   := 23;
      Left     := FilterControls[i].Comparison.Left + FilterControls[i].Comparison.Width + 28;
      Top      := 24 + 40 * i;
      Parent   := tabControl;
      AutoSize := False;
      Hint     := 'Enter the quantity to include with this filter';
      ShowHint := True;
      TabOrder := 5 * i + 4;
      Tag      := i;
    end;

    FilterControls[i].AndOr := Tcombobox.Create(nil);
    with FilterControls[i].AndOr do
    begin
      Visible  := False;
      Width    := 64;
      Height   := 23;
      Left     := FilterControls[i].Quantity.Left + FilterControls[i].Quantity.Width + 28;
      Top      := 24 + 40 * i;
      Parent   := tabControl;
      Style    := csDropDownList;
      Hint     := 'Select ''END'' to finish the last line of the filter';
      ShowHint := True;
      TabOrder := 5 * i + 5;
      Tag      := i;
      OnChange := @AndOrOnChange;
      Items.Add('And');
      Items.Add('End');
      ItemIndex := 1;
    end;
  end;
end;

procedure TvFilter.CreateOptionsControls;
begin
  with OptionsControl do begin
    Caption  := 'Sticky Filter';
    ShowHint := True;
    Hint     := 'Use the same filter for all tables';
    Height   := 74;
    Width    := 135;
  end;

  chkOptionCourse := TcheckBox.Create(nil);
  with chkOptionCourse do begin
    Parent   := OptionsControl;
    checked  := True;
    Caption  := 'include Course';
    Top      := 5;
    Left     := 5;
    ShowHint := True;
    Hint     := Parent.Hint;
    OnClick  := @OptionsOnChange;
  end;

  chkOptionDates := TcheckBox.Create(nil);
  with chkOptionDates do begin
    Parent   := OptionsControl;
    checked  := True;
    Caption  := 'include Dates';
    Top      := 27;
    Left     := 5;
    ShowHint := True;
    Hint     := Parent.Hint;
    OnClick  := @OptionsOnChange;
  end;

end;

procedure TvFilter.CreateWhere;
var
  Where: string = '';
begin
  AddDatesFilter(Where);
  AddCourseFilter(Where);
  AddTableSpecificFilter(Where);

  StringData.LoadTable(CurrentTable, Where);
end;

procedure TvFilter.CouresOnChange(Sender: Tobject);
begin
  chkCourses.Checked := True;
  FilterControlOnChange(Sender);
end;

procedure TvFilter.FilterControlOnChange(Sender: Tobject);
begin
  CreateWhere;
  Graphs.DrawGraph(False);
end;

procedure TvFilter.DateFromOnChange(Sender: Tobject; var Adate: TDateTime;
  var AcceptDate: boolean);
begin
  chkDateFrom.Checked := True;
  TdateFrom           := DateToStr(Adate);
  FilterControlOnChange(Sender);
  Unused(AcceptDate);
end;

procedure TvFilter.DateToOnChange(Sender: Tobject; var Adate: TDateTime;
  var AcceptDate: boolean);
begin
  chkDateTo.Checked := True;
  TdateTo           := DateToStr(Adate);
  FilterControlOnChange(Sender);
  Unused(AcceptDate);
end;

constructor TvFilter.Create(QuickCntrl, TabCntrl, OptionsCntrl: TwinControl;Corses: TvCourses;
    StrData: TvStringData; Tdata: TvTableData; FilterINI: TvINI; Grphs: TvGraphs);
begin
  CurrentWhere          := '';
  QuickFilterControl    := QuickCntrl;
  TabControl            := TabCntrl;
  OptionsControl        := OptionsCntrl;
  Courses               := Corses;
  StringData            := StrData;
  TableData             := Tdata;
  Fini                  := FilterINI;
  Graphs                := Grphs;
  TdateFrom             := FormatDateTime('dd/mm/yyyy', Date-187);
  TdateTo               := FormatDateTime('dd/mm/yyyy', Date);
  CreateControls;
  RestoreOptions;
end;

destructor TvFilter.Destroy;
begin
  Courses.unRegisterComboBox(cmbCourses);
  cmbCourses.Destroy;
  chkCourses.Destroy;
  chkDateFrom.Destroy;
  chkDateTo.Destroy;
  dteDateFrom.Destroy;
  dteDateTo.Destroy;
  lblCourses.Destroy;
  lblDateFrom.Destroy;
  lblDateTo.Destroy;
  cmbFilter.Destroy;
  btnUseFilter.Destroy;
  btnDeleteFilter.Destroy;
  btnSaveFilter.Destroy;
  edtFilter.Destroy;
  Label1.Destroy;
  btnFilter.Destroy;
  gbxFilter.Destroy;
  chkOptionCourse.Destroy;
  chkOptionDates.Destroy;
  DestroyFilterControls;
  inherited Destroy;
end;

procedure TvFilter.DisplayQuickFilters;
var
  Display: boolean;
begin
  if not Options.StickyCourse then
    chkCourses.Checked  := False;
  if not Options.StickyDates then begin
    chkDateFrom.Checked := False;
    chkDateTo.Checked   := False;
  end;

  Display             := Tabledata.IndexOfField(vCourse,CurrentTable) > -1;
  cmbCourses.Visible  := Display;
  chkCourses.Visible  := Display;
  lblCourses.Visible  := Display;

  Display             := Tabledata.IndexOfField(vDate,CurrentTable) > -1;
  chkDateFrom.Visible := Display;
  chkDateTo.Visible   := Display;
  dteDateFrom.Visible := Display;
  dteDateTo.Visible   := Display;
  lblDateFrom.Visible := Display;
  lblDateto.Visible   := Display;
end;

procedure TvFilter.ActivateFilter;
begin
  cmbCourses.onSelect      := @CouresOnChange;
  dteDateFrom.OnAcceptDate := @DateFromOnChange;
  dteDateTo.OnAcceptDate   := @DateToOnChange;
end;

procedure TvFilter.AddCourseFilter(var Where: string);
var
  i: integer;
begin
  i := TableData.IndexOfField(vCourse, CurrentTable);
  if chkCourses.Checked  and (i > -1) then
  begin
    Where := ifthen(Where = '',Where, Where + ' and ');
    Where := Where + ' ' + TableData.IthFieldName(i, CurrentTable) + ' = ' +
      Courses.CourseID(cmbCourses.Items[cmbCourses.ItemIndex]);
  end;
end;

procedure TvFilter.AddDatesFilter(var Where: string);
var
  i: integer;
begin
  i := TableData.IndexOfField(vDate, CurrentTable);
  if i > -1 then begin
    if chkDateFrom.Checked then
    begin
      Where := ifthen(Where = '', Where, Where + ' and ');
      Where := Where + ' ' + TableData.IthFieldName(i, CurrentTable) + ' >= ''' +
        FormatDate(TdateFrom) + '''';
    end;

    if chkDateTo.Checked then
    begin
      Where := ifthen(Where = '', Where, Where + ' and ');
      Where := Where + ' ' + TableData.IthFieldName(i, CurrentTable) + ' <= ''' +
        FormatDate(TdateTo) + '''';
    end;
  end;
end;

procedure TvFilter.AddTableSpecificFilter(var Where: string);
var
  i: integer;
begin
  for i := 0 to MaxFilterControls do
    with FilterControls[i] do
    begin
      if not enabled.Visible then break;
      if (not Enabled.Checked) or (Quantity.Text = '') then continue;
      Where := ifthen(Where = '', Where, Where + ' and ');
      if CompareText(Comparison.Text, 'contains') = 0 then
      begin
        if TableData.GetFieldType(Field.Text, CurrentTable) = vSeconds then
          Where := Where + TableData.GetFieldName(Field.Text, CurrentTable) +
            ' like ''%' + FormatSeconds(Quantity.Text) + '%'''
        else
          Where := Where + TableData.GetFieldName(Field.Text, CurrentTable) +
            ' like ''%' + Quantity.Text + '%''';
      end
      else if CompareText(Comparison.Text, 'excludes') = 0 then
      begin
        if TableData.GetFieldType(Field.Text, CurrentTable) = vSeconds then
          Where := Where + ' not (' + TableData.GetFieldName(Field.Text, CurrentTable) +
            ' like ''%' + FormatSeconds(Quantity.Text) + '%'')'
        else
          Where := Where + ' not (' + TableData.GetFieldName(Field.Text, CurrentTable) +
            ' like ''%' + Quantity.Text + '%'')';
      end
      else if CompareText(Comparison.Text, 'not =') = 0 then
      begin
        if TableData.GetFieldType(Field.Text, CurrentTable) = vSeconds then
          Where := Where + ' not (' + TableData.GetFieldName(Field.Text, CurrentTable) +
            ' = ''' + FormatSeconds(Quantity.Text) + ''')'
        else
          Where := Where + ' not (' + TableData.GetFieldName(Field.Text, CurrentTable) +
            ' = ''' + Quantity.Text + ''')';
      end
      else
      begin
        if TableData.GetFieldType(Field.Text, CurrentTable) = vSeconds then
          Where := Where + TableData.GetFieldName(Field.Text, CurrentTable) +
            ' ' + Comparison.Text + ' ''' + FormatSeconds(Quantity.Text) + ''''
        else
          Where := Where + TableData.GetFieldName(Field.Text, CurrentTable) +
            ' ' + Comparison.Text + ' ''' + Quantity.Text + '''';
      end;
    end;
end;

procedure TvFilter.LoadTable(Table: string);
begin
  CurrentTable := Table;
  DisplayQuickFilters;
  CreateWhere;
  Graphs.DrawGraph;
end;

procedure TvFilter.DestroyFilterControls;
var
  i: integer;
begin
  for i := 0 to MaxFilterControls do
    with FilterControls[i] do begin
      AndOr.Destroy;
      Comparison.Destroy;
      Enabled.Destroy;
      Quantity.Destroy;
      Field.Destroy;
    end;
end;

procedure TvFilter.InitializeTableSpecificControls;
var
  i: integer;
begin
  for i := 0 to MaxFilterControls do
    with FilterControls[i] do begin
      AndOr.Visible        := false;
      AndOr.ItemIndex      := 1;
      Comparison.Visible   := false;
      Comparison.ItemIndex := 0;
      Enabled.Visible      := false;
      Enabled.Checked      := False;
      Quantity.Visible     := false;
      Quantity.Text        := '';
      Field.Visible        := false;
      Field.Items.Clear;
    end;
end;

procedure TvFilter.AddFilterControlRow(Row:integer);
begin
  if Row > MaxFilterControls then exit;

  with FilterControls[Row] do
  begin
    Enabled.Visible    := True;
    Enabled.Checked    := True;
    Field.Visible      := True;
    Comparison.Visible := True;
    Quantity.Visible   := True;
    AndOr.Visible      := True;
  end;

  FillTableSpecificControls(Row);
end;

procedure TvFilter.FillTableSpecificControls(CurrFC:integer);
begin
  TableData.VisibleUFfieldNames(FilterControls[CurrFC].Field, CurrentTable);
  edtFilter.Text := 'Filter 1';
end;

procedure TvFilter.AndOrOnChange(Sender: TObject);
var
  cmb: TComboBox;
begin
  cmb := Sender as TComboBox;
  if cmb.tag = MaxFilterControls then begin
    cmb.ItemIndex := 1;
    exit;
  end;
  if (CompareText(cmb.Text, 'End') <> 0) and (not FilterControls[cmb.tag + 1].Enabled.Visible) then
    AddFilterControlRow(cmb.tag+1);
end;

procedure TvFilter.LoadFilters;
begin
  btnUseFilter.Enabled := False;
  btnDeleteFilter.Enabled := False;
  Fini.LoadFilters(CurrentTable, cmbFilter);
  if cmbFilter.Items.Count > 0 then
  begin
    cmbFilter.ItemIndex := 0;
    btnUseFilter.Enabled := True;
    btnDeleteFilter.Enabled := True;
  end;
end;

procedure TvFilter.btnUseFilterClick(Sender: TObject);
var
  i: integer = 0;
  FilterRow: vFilterRow;
begin
  InitializeTableSpecificControls;
  while i <= MaxFilterControls do
    with FilterControls[i] do
    begin
      if not Fini.AfilterRow(i, cmbFilter.Text, FilterRow) then break;
      AddFilterControlRow(i);
      Enabled.Checked      := True;
      Field.ItemIndex      := Field.Items.IndexOf(TableData.GetUFfieldName(FilterRow.Field));
      Comparison.ItemIndex := Comparison.Items.IndexOf(FilterRow.Comparison);
      Quantity.Text        := FilterRow.Quantity;
      AndOr.OnChange       := nil;
      AndOr.ItemIndex      := 0;
      AndOr.OnChange       := @AndOrOnChange;
      i                    := i + 1;
    end;
  if i = 0 then
    AddFilterControlRow(i);
  if i > 0 then
    FilterControls[i - 1].AndOr.ItemIndex := 1;
  if FilterRow.Course <> -1 then
  begin
    cmbCourses.ItemIndex := cmbCourses.Items.IndexOf(Courses.CourseName(FilterRow.Course));
    chkCourses.Checked := True;
  end
  else
    chkCourses.Checked := False;
  if FilterRow.DateFrom <> '' then
  begin
    dteDateFrom.Text    := FilterRow.DateFrom;
    TdateFrom           := FilterRow.DateFrom;
    chkDateFrom.Checked := True;
  end
  else
    chkDateFrom.Checked := False;
  if FilterRow.DateTo <> '' then
  begin
    dteDateTo.Text    := FilterRow.DateTo;
    TdateTo           := FilterRow.DateTo;
    chkDateTo.Checked := True;
  end
  else
    chkDateTo.Checked := False;
  edtFilter.Text := cmbFilter.Text;

  btnFilterClick(Sender);
end;

procedure TvFilter.btnSaveFilterClick(Sender: TObject);
begin
  if edtFilter.Text = '' then
  begin
    MessageDlg('Warning -- Filter must have a name',
      'Please provide a name for the filter.', mtWarning, [mbOK], 0);
    exit;
  end;
  if  Fini.FilterExists(edtFilter.Text) and (MessageDlg('Warning -- Filter about to be overwritten',
      'The filter ''' + edtFilter.Text + ''' already exists. Do you wish to overwrite?',
      mtWarning, mbYesNo, 0) = mrNo) then
      exit;
  if not SaveFilter then
  begin
    MessageDlg('Warning -- No conditions to filter.',
      'There is nothing to save. Please complete at least the first line of the filter.',
      mtWarning, [mbOK], 0);
    exit;
  end;
  btnFilterClick(Sender);
end;

procedure TvFilter.btnFilterClick(Sender: TObject);
begin
  FilterControlOnChange(Sender);
  (TabControl.Parent as TPageControl).ActivePage := (TabControl.Parent as TPageControl).Pages[0];
end;

procedure TvFilter.btnDeleteFilterClick(Sender: TObject);
begin
  if MessageDlg('Warning -- Filter about to be deleted',
    'Are you sure you wish to delete the filter ''' + cmbFilter.Text + '''?',mtWarning,
    mbYesNo,0) = mrNo then
    exit;
  Fini.DeleteFilter(cmbFilter.Text);
  cmbFilter.Items.Delete(cmbFilter.Items.IndexOf(cmbFilter.Text));
  if cmbFilter.Items.Count > 0 then
    cmbFilter.ItemIndex := 0
  else
  begin
    btnDeleteFilter.Enabled := False;
    btnUseFilter.Enabled := False;
  end;
end;

function TvFilter.SaveFilter: boolean;
var
  i: integer = 0;
  SomethingSaved: boolean = False;
  FilterRow: vFilterRow = (
    Quantity:   '';
    Field:      '';
    Comparison: '';
    Course:     -1;
    DateFrom:   '';
    DateTo:     '';
  );
begin
  if chkCourses.Checked or chkDateFrom.Checked or chkDateTo.Checked then
    SomethingSaved     := True;
  if chkCourses.Checked then
    FilterRow.Course   := StrToInt(Courses.CourseID(cmbCourses.Text));
  if chkDateFrom.Checked then
    FilterRow.DateFrom := dteDateFrom.Text;
  if chkDateTo.Checked then
    FilterRow.DateTo   := dteDateTo.Text;
  while i <= MaxFilterControls do
    with FilterControls[i] do
    begin
      if not Enabled.Visible then break;
      if Enabled.Checked and (Quantity.Text <> '') then
      begin
        SomethingSaved       := True;
        FilterRow.Field      := TableData.GetFieldName(Field.Text);
        FilterRow.Comparison := Comparison.Text;
        FilterRow.Quantity   := Quantity.Text;
        Fini.SaveFilterRow(i,edtFilter.Text,CurrentTable, FilterRow);
      end;
      i := i + 1;
    end;
  if SomethingSaved then
  begin
    if cmbFilter.Items.IndexOf(edtFilter.Text) < 0 then
      cmbFilter.Items.Add(edtFilter.Text);
    cmbFilter.ItemIndex     := cmbFilter.Items.IndexOf(edtFilter.Text);
    btnDeleteFilter.Enabled := True;
    btnUseFilter.Enabled    := True;
  end;
  Result := SomethingSaved;
end;

procedure TvFilter.OptionsOnChange(Sender: TObject);
begin
  Options.StickyDates  := chkOptionDates.Checked;
  Options.StickyCourse := chkOptionCourse.Checked;
  SaveOptions;
end;

procedure TvFilter.SaveOptions;
begin
  Fini.SaveFilterOptions(Options);
end;

procedure TvFilter.RestoreOptions;
begin
  Fini.RestoreFilterOptions(Options);
  chkOptionDates.OnClick  := nil;
  chkOptionCourse.OnClick := nil;
  chkOptionDates.Checked  := Options.StickyDates;
  chkOptionCourse.Checked := Options.StickyCourse;
  chkOptionDates.OnClick  := @OptionsOnChange;
  chkOptionCourse.OnClick := @OptionsOnChange;
end;

end.

