unit vGraphs;

/////////////////////////////////////////////////////////////////////
////                             vGraphs                         ////
////                                                             ////
////   - Unit to implement the drawing of statistical graphs     ////
////                                                             ////
////   Notes:                                                    ////
////   - Yvals[i] contains metadata on each field:               ////
////     - whether to graph this field                           ////
////     - the type of graph (bar or line graph)                 ////
////     - the X axis to use (left (0) or right (2))             ////
////     - the Series to use to store it's data                  ////
////   - Yvals[i] is evaluated in 'DetermineGraphType            ////
////     - A field measured in seconds (eg duration of ride)     ////
////       is always graphed on the left axis                    ////
////     - in a Mixed Graph (a mixture of line and bar graphs)   ////
////       the left axis uses the line graph, while all other    ////
////       series are graphed as bar graphs                      ////
////                                                             ////
////   - After determining the graph type,                       ////
////     'ActivateSeriesForAppropriateAxis' activates all series ////
////     assigned to Yvals[i]:                                   ////
////   - and assigns the series to the correct axis, sets        ////
////     the title of the axis, sets the colour, and width of    ////
////     the bar in a bar series                                 ////
////                                                             ////
////   - Finally, 'AddDataPoints' iterates through the           ////
////     StringData adding the data to the Source of each series ////
////                                                             ////
////   - There are 4 Line series created and 4 bar series        ////
////     created. That is, no more than 4 items can be graphed.  ////
////                                                             ////
////   - A random colour is assigned to each Field, but can be   ////
////     changed by the user.                                    ////
////                                                             ////
/////////////////////////////////////////////////////////////////////

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, ComCtrls, Dialogs, TAgraph, TASeries, TAchartAxis, TAchartutils, TAchartAxisutils, TATransformations, TAsources,
  graphics, ExtCtrls, StdCtrls, vTypes, LCLtype, vStringData, vTableData, vUtils, TAtypes;

const
  MaxYchkBoxes = 10; // Max number of checkboxes that will fit on the screen.

type

  { TvGraphs }

  // Type of Graphs
  vGraphTypes = (vNoGraph, vLineGraph, vBarGraph, vMixedGraph, vPointGraph);

  TvGraphs = class
    private
      StringData: TvStringData;
      TableData:  TvTableData;
      TabControl: TtabSheet;
      chtResults: TChart;
      chtResultsBarSeries1: TBarSeries;
      chtResultsBarSeries2: TBarSeries;
      chtResultsBarSeries3: TBarSeries;
      chtResultsBarSeries4: TBarSeries;
      chtResultsLineSeries1: TLineSeries;
      chtResultsLineSeries2: TLineSeries;
      chtResultsLineSeries3: TLineSeries;
      chtResultsLineSeries4: TLineSeries;
      catResultsLeft: TchartAxisTransformations;
      catResultsRight: TchartAxisTransformations;
      catResultsRightAutoTransform: TautoScaleAxisTransform;
      catResultsLeftAutoTransform: TautoScaleAxisTransform;
      ListChartSource1: TlistChartSource;
      ListChartSource2: TlistChartSource;
      ListChartSource3: TlistChartSource;
      ListChartSource4: TlistChartSource;
      rdgGraphType: TradioGroup;
      chkGridLinesLeftAxis: TcheckBox;
      chkGridLinesRightAxis: TcheckBox;
      chkSeparateGraph: TcheckBox;
      // The index to the current X and Y values being graphed. Needed for processing in the MarkToText procedure
      CurrPlottedYval: integer;
      CurrPlottedXval: integer;
      GraphType: vGraphTypes;
      // Array of checkboxes for user to select Y values for Graph
      Yvals: array[0..MaxFields] of record
        GraphType: vGraphTypes;
        Yaxis,Series: integer;
        Enabled: Tcheckbox;
        Colour: TShape;
        Xval: TRadioButton;
      end;
      // Heading for legend (Yvals)
      lblYval: Tlabel;
      lblXval: Tlabel;
      // Current table being graphed
      CurrentTable: string;
      ///// >>>>> Procedures <<<<< /////
      procedure CreateControls;
      procedure CreateGraphControls;
      procedure CreateOtherControls;
      procedure CreateGraph;
      procedure ConfigureChart;
      // 4 bar and line series are associated with chtResults
      // the series are actived and deactivated as per the user's requests.
      procedure AssociateSeriesToChart;
      procedure ConfigureSeries;
      procedure ConfigureLineSeries(var Series: TLineSeries);
      procedure ConfigureBarSeries(var Series: TBarSeries);
      procedure AssociateTransformations;
      procedure ConfigureTransformations;
      procedure ConfigureRdgGraphType;
      procedure ConfigureChkGridLines;
      procedure ConfigureChkSeparateGraph;
      procedure CreateYvalsArray;
      // Turns on/off grid lines for axis 0, then axis 2 and then both
      procedure chkGridLinesChange(Sender: TObject);
      // split the left and right axes data into top and bottom of graph
      procedure chkSeparateGraphChange(Sender: TObject);
      // Change graph type to Line, Bar or a mixture
      procedure rdgGraphTypeClick(Sender: TObject);
      // Clear the settings on the CheckBoxes (Yvals[i].enabled, etc)
      procedure DestroyGraphCheckboxes;
      // Draw the CheckBoxes (Yvals) for the non-hidden fields
      procedure DrawGraphCheckBoxes;
      // Clear or deactivate all series
      procedure ClearSeries();
      // Check CheckBox has been ticked, clear existing series, etc.
      function PrepareGraph: boolean;
      // Return the number of Graph checkboxes checked, and the index of the SecondsField (or -1)
      function NumberOfCheckedCheckBoxes(out SecondsField:integer): integer;
      // Determine the type of graph and the axis for each selected CheckBox (Yval)
      procedure DetermineGraphTypes(SecondsField:integer);
      // Activate and configure series (eg axis title, barwidth, etc)
      procedure ActivateSeriesForAppropriateAxis(const NumberChecked: integer);
      procedure ConfigureBarSeries(const i, NumberChecked: integer);
      procedure ConfigureLineSeries(const i: integer);
      // If right axis is visible, show the Right Axis Grid Lines Check Box
      procedure ShowRightGridLinesCheckBox;
      // On click of a checkbox, redraw the graph
      procedure YvalsClick(Sender: TObject);
      // On click of a Xval checkbox, uncheck all other checkboxes, and redraw the graph
      procedure XvalsClick(Sender: TObject);
      // Add a Main title to the graph
      procedure AddTopBottomTitles;
      // Allow user to change the colour of the graph's series
      procedure YvalsColourOnClick(Sender: TObject);
      // Format Seconds field if required
      procedure chtResultsAxisList0MarkToText(var AText: string; AMark: double);
      // Add the data values to the graph.
      procedure AddDataPoints;
      // Format Bottom Axis Data
      function FormatBottomAxisDataPoint(const Row: integer): double;
  public
      constructor Create(TabCtrl: TtabSheet; StrData: TvStringData; Tdata: TvTableData);
      destructor Destroy(); override;
      // Draw the graph!
      procedure DrawGraph(const ResetYvalues: boolean = True);
  end;

implementation

{ TvGraphs }

uses StrUtils;

procedure TvGraphs.CreateControls;
begin
  CreateGraph;
  CreateOtherControls;
end;

procedure TvGraphs.CreateGraph;
begin
  CreateGraphControls;
  ConfigureChart;
  AssociateSeriesToChart;
  ConfigureSeries;
  AssociateTransformations;
  ConfigureTransformations;
end;

procedure TvGraphs.CreateOtherControls;
begin
  rdgGraphType := TradioGroup.Create(nil);
  ConfigureRdgGraphType;
  chkGridLinesLeftAxis  := TcheckBox.Create(nil);
  chkGridLinesRightAxis := TcheckBox.Create(nil);
  ConfigureChkGridLines;
  chkSeparateGraph := TcheckBox.Create(nil);
  ConfigureChkSeparateGraph;
  CreateYvalsArray;
end;

constructor TvGraphs.Create(TabCtrl: TtabSheet; StrData: TvStringData; Tdata: TvTableData);
begin
  TabControl   := TabCtrl;
  StringData   := StrData;
  TableData    := Tdata;
  CurrentTable := TableData.TableName;
  GraphType    := vLineGraph;
  StringData.RegisterDrawGraph(@DrawGraph);
  CreateControls;
end;

destructor TvGraphs.Destroy();
var
  i: integer;
begin
  chtResultsBarSeries1.Destroy;
  chtResultsBarSeries2.Destroy;
  chtResultsBarSeries3.Destroy;
  chtResultsBarSeries4.Destroy;
  chtResultsLineSeries1.Destroy;
  chtResultsLineSeries2.Destroy;
  chtResultsLineSeries3.Destroy;
  chtResultsLineSeries4.Destroy;
  catResultsRightAutoTransform.Destroy;
  catResultsLeftAutoTransform.Destroy;
  catResultsLeft.Destroy;
  catResultsRight.Destroy;
  ListChartSource1.Destroy;
  ListChartSource2.Destroy;
  ListChartSource3.Destroy;
  ListChartSource4.Destroy;
  rdgGraphType.Destroy;
  chkGridLinesLeftAxis.Destroy;
  chkGridLinesRightAxis.Destroy;
  chkSeparateGraph.Destroy;
  chtResults.Destroy;
  lblYval.Destroy;
  lblXval.Destroy;
  for i := 0 to MaxFields do
    with Yvals[i] do begin
      Enabled.Free;
      Colour.Free;
    end;
  inherited destroy;
end;

procedure TvGraphs.DrawGraph(const ResetYvalues: boolean = True);
var
  NumberChecked, SecondsField: integer;
begin
  if not TableData.GetGraphed() then exit;
  if ResetYvalues then DrawGraphCheckBoxes;
  if not PrepareGraph then exit;
  NumberChecked := NumberOfCheckedCheckBoxes(SecondsField);
  DetermineGraphTypes(SecondsField);
  ActivateSeriesForAppropriateAxis(NumberChecked);
  AddDataPoints;
end;

procedure TvGraphs.chkGridLinesChange(Sender: TObject);
begin
  chtResults.AxisList.Axes[0].Grid.Visible := chkGridLinesLeftAxis.Checked;
  chtResults.AxisList.Axes[2].Grid.Visible := chkGridLinesRightAxis.Checked;
end;

procedure TvGraphs.chkSeparateGraphChange(Sender: TObject);
begin
  if chkSeparateGraph.Checked then begin
    catResultsRightAutoTransform.MinValue:=1;
    catResultsRightAutoTransform.MaxValue:=2;
    chtResults.AxisList.Axes[1].Grid.Visible:=True;
  end
  else begin
    catResultsRightAutoTransform.MinValue:=0;
    catResultsRightAutoTransform.MaxValue:=1;
    chtResults.AxisList.Axes[1].Grid.Visible:=False;
  end;
end;

procedure TvGraphs.chtResultsAxisList0MarkToText(var AText: string; AMark: double);
begin
  if TableData.GetFieldType(CurrPlottedYval) = vSeconds then
    AText := FormatSeconds(FloatToStr(Amark), True);
end;

procedure TvGraphs.rdgGraphTypeClick(Sender: TObject);
begin
  case rdgGraphType.ItemIndex of
    0: GraphType := vLineGraph;
    1: GraphType := vPointGraph;
    2: GraphType := vMixedGraph;
    3: GraphType := vBarGraph;
  end;
  DrawGraph(False);
end;

procedure TvGraphs.DestroyGraphCheckboxes;
var
  i: integer;
begin
  for i := 0 to MaxFields do begin
    YVals[i].Enabled.visible :=false;
    YVals[i].Xval.visible    :=false;
    YVals[i].Colour.visible  :=false;
    Yvals[i].Enabled.OnClick := nil;
    Yvals[i].Enabled.Checked := False;
    Yvals[i].GraphType       := vNoGraph;
    Yvals[i].Yaxis           := -1;
    Yvals[i].Series          := -1;
  end;
end;

procedure TvGraphs.DrawGraphCheckBoxes;
var
  i: integer;
  j:integer = -1;
  Tstr: string;
begin
  DestroyGraphCheckboxes;
  for i := 1 to TableData.GetNumberOfFields(CurrentTable,False)  do
    if not ((TableData.GetFieldType(i) in [vString, vDescription,
      {vDate,} vCourse, vURL]) or  TableData.GetFieldHidden(i)) then
    begin
      j:=j+1;
      if j = MaxYchkBoxes then break;
      if i = TableData.IndexOfDefYvalue then Yvals[j].Enabled.Checked := True;
      if i = TableData.IndexOfDefXvalue then Yvals[j].Xval.Checked    := True;
      Tstr := TableData.GetithUFfieldName(i);
      Yvals[j].Enabled.Caption := ifthen(Length(Tstr) < 15, Tstr, LeftStr(Tstr,14) + '...');
      Yvals[j].Enabled.Tag     := i;
      Yvals[j].Xval.Tag        := i;
      Yvals[j].Enabled.Visible := True;
      Yvals[j].Xval.Visible    := True;
      Yvals[j].Colour.Visible  := True;
      Yvals[j].Enabled.OnClick := @YvalsClick;
      Yvals[j].Xval.OnClick    := @XvalsClick;
      if TableData.GetFieldType(i) = vDate then begin
        Yvals[j].Enabled.Enabled := False;
        Yvals[j].Colour.Visible  := False;
      end;
    end;
    CurrPlottedXval := TableData.IndexOfDefXvalue;
end;

function TvGraphs.PrepareGraph: boolean;
var
  i: integer;
begin
  Result := False;
  if NumberOfCheckedCheckBoxes(i) = 0 then
  begin
    MessageDlg('Warning -- Incorrect number selected.',
      'At least one must be selected.', mtWarning, [mbOK], 0);
    exit;
  end;
  if NumberOfCheckedCheckBoxes(i) > 4 then
  begin
    MessageDlg('Warning -- Incorrect number selected.',
      'No more than 4 may be selected.', mtWarning, [mbOK], 0);
    exit;
  end;

  // Clear existing series etc.
  ClearSeries();

  // Draw Grids
  chkGridLinesChange(nil);
  chkSeparateGraphChange(nil);

  // Setting the colour in ConfigureChart doesn't work, so it is set here.
  chtResults.Color := clCream;
  Result           := True;
end;

function TvGraphs.NumberOfCheckedCheckBoxes(out SecondsField:integer): integer;
var
  i: integer;
  j: integer = 0;
begin
  SecondsField := -1;
  for i := 0 to (MaxFields) do
    if Yvals[i].Enabled.Checked then begin
      j := j + 1;
      if TableData.GetFieldType(Yvals[i].Enabled.Tag) = vSeconds then
        SecondsField := i;
    end;
  Result := j;
end;

procedure TvGraphs.ClearSeries();
var
  i:integer;
begin
  for i := 0 to chtResults.SeriesCount - 1 do
    chtResults.Series.Items[i].Active:=False;

  for i := 0 to 3 do
    (chtResults.Series.Items[i] as TBarSeries).ListSource.Clear;

  chtResults.LeftAxis.Title.Caption := '';
  chtResults.BottomAxis.Title.Caption := '';
  chtResults.AxisList.Axes[2].Title.Caption := '';
  chtResults.Title.Text.Clear;

  chtResults.AxisList.Axes[0].onMarkToText := @chtResultsAxisList0MarkToText;
  chtResults.AxisList.Axes[2].Visible:=False;
  chtResults.AxisList.Axes[2].Title.Visible:=False;
end;

procedure TvGraphs.DetermineGraphTypes(SecondsField:integer);
var
  i:integer;
  first:integer = -1;
  series: integer = 0; // assume a bar series (bar series = 0 - 3, line series = 4 - 7)
begin
  if GraphType in [vLineGraph,vPointGraph] then series := 4;
  for i := 0 to MaxFields do begin
    if not Yvals[i].Enabled.Visible then break;
    if Yvals[i].enabled.Checked then begin
      case GraphType of
        vLineGraph,vBarGraph, vPointGraph: begin
                                Yvals[i].GraphType:=GraphType;
                                Yvals[i].Series:=series;
                                Yvals[i].Yaxis:=2;
                                if first = -1 then begin
                                  first := i;
                                  Yvals[i].Yaxis:=0;
                                end
                                else if i = SecondsField then begin
                                  Yvals[i].Yaxis:=0;
                                  Yvals[first].Yaxis:=2;
                                end;
                                if GraphType = vPointGraph then begin
                                  (chtResults.Series.Items[Series] as TLineSeries).LineType   := ltNone;
                                  (chtResults.Series.Items[Series] as TLineSeries).ShowPoints := True;
                                end
                                else if GraphType = vLineGraph then begin
                                  (chtResults.Series.Items[Series] as TLineSeries).LineType   := ltFromPrevious;
                                  (chtResults.Series.Items[Series] as TLineSeries).ShowPoints := False;
                                end;
                              end;
        vMixedGraph: begin
                       Yvals[i].GraphType:=vBarGraph;
                       Yvals[i].Yaxis:=2;
                       Yvals[i].Series:=series;
                       if first = -1 then begin
                         Yvals[i].GraphType:=vLineGraph;
                         first := i;
                         Yvals[i].Yaxis:=0;
                         Yvals[i].Series:=4;
                       end
                       else if i = SecondsField then begin
                         Yvals[i].GraphType:=vLineGraph;
                         Yvals[i].Yaxis:=0;
                         Yvals[i].Series:=series+4;
                         Yvals[first].GraphType:=vBarGraph;
                         Yvals[first].Yaxis:=2;
                         Yvals[first].Series:=0;
                       end;
                     end;
      end;
      series := series + 1;
    end;
  end;
end;

procedure TvGraphs.CreateGraphControls;
begin
  chtResultsBarSeries1         := TbarSeries.Create(nil);
  chtResultsBarSeries2         := TbarSeries.Create(nil);
  chtResultsBarSeries3         := TbarSeries.Create(nil);
  chtResultsBarSeries4         := TbarSeries.Create(nil);
  chtResultsLineSeries1        := TlineSeries.Create(nil);
  chtResultsLineSeries2        := TlineSeries.Create(nil);
  chtResultsLineSeries3        := TlineSeries.Create(nil);
  chtResultsLineSeries4        := TlineSeries.Create(nil);
  catResultsLeft               := TchartAxisTransformations.Create(nil);
  catResultsRight              := TchartAxisTransformations.Create(nil);
  catResultsRightAutoTransform := TautoScaleAxisTransform.Create(nil);
  catResultsLeftAutoTransform  := TautoScaleAxisTransform.Create(nil);
  ListChartSource1             := TlistChartSource.Create(nil);
  ListChartSource2             := TlistChartSource.Create(nil);
  ListChartSource3             := TlistChartSource.Create(nil);
  ListChartSource4             := TlistChartSource.Create(nil);
  chtResults                   := Tchart.Create(nil);
  lblYval                      := Tlabel.Create(nil);
  lblXval                      := Tlabel.Create(nil);
end;

procedure TvGraphs.ConfigureChart;
begin
  with chtResults do begin
    Parent    := TabControl;
    Left      := 4;
    Top       := 4;
    Width     := 648;
    Height    := 288;
    BackColor := clWhite;
    With BottomAxis do begin
      Grid.Visible := False;
      with Marks do begin
        Format        := '%2:s';
        Style         := smsLabel;
        OverlapPolicy := opHideNeighbour;
        with LabelFont do begin
          Orientation := 360;
          Style       := [fsBold];
          Size        := 10;
        end;
      end;
    end;
    With Frame do begin
      Color     := clBlack;
      Visible   := True;
      Width     := 2;
    end;
    With LeftAxis do begin
      Alignment     := calLeft;
      Visible       := True;
      Grid.Visible  := False;
      Marks.Format  := '%0:.9g';
      Title.Visible := true;
    end;
    With Title do begin
      Visible   := True;
      Font.Size := 16;
    end;
    With TchartAxis.Create(AxisList) do begin
      Alignment                   := calRight;
      Visible                     := True;
      Grid.Visible                := False;
      Title.LabelFont.Orientation := 900;
    end;
  end;
  with lblYval do begin
    top        := chtResults.Top;
    left       := chtResults.Left+chtResults.Width;
    caption    := 'Y axis';
    font.Color := clBlue;
    font.Style := [fsBold];
    parent     := TabControl;
  end;
  with lblXval do begin
    top        := chtResults.Top;
    left       := chtResults.Left+chtResults.Width+136;
    caption    := 'X axis';
    font.Color := clBlue;
    font.Style := [fsBold];
    parent     := TabControl;
  end;
end;

procedure TvGraphs.AssociateSeriesToChart;
begin
  with ChtResults do begin
    AddSeries(chtResultsBarSeries1);
    AddSeries(chtResultsBarSeries2);
    AddSeries(chtResultsBarSeries3);
    AddSeries(chtResultsBarSeries4);
    AddSeries(chtResultsLineSeries1);
    AddSeries(chtResultsLineSeries2);
    AddSeries(chtResultsLineSeries3);
    AddSeries(chtResultsLineSeries4);
  end;
end;

procedure TvGraphs.ConfigureSeries;
begin
  ConfigureBarSeries(chtResultsBarSeries1);
  chtResultsBarSeries1.Source := ListChartSource1;
  ConfigureBarSeries(chtResultsBarSeries2);
  chtResultsBarSeries2.Source := ListChartSource2;
  ConfigureBarSeries(chtResultsBarSeries3);
  chtResultsBarSeries3.Source := ListChartSource3;
  ConfigureBarSeries(chtResultsBarSeries4);
  chtResultsBarSeries4.Source := ListChartSource4;
  ConfigureLineSeries(chtResultsLineSeries1);
  chtResultsLineSeries1.Source := ListChartSource1;
  ConfigureLineSeries(chtResultsLineSeries2);
  chtResultsLineSeries2.Source := ListChartSource2;
  ConfigureLineSeries(chtResultsLineSeries3);
  chtResultsLineSeries3.Source := ListChartSource3;
  ConfigureLineSeries(chtResultsLineSeries4);
  chtResultsLineSeries4.Source := ListChartSource4;
end;

procedure TvGraphs.ConfigureLineSeries(var Series: TlineSeries);
begin
  with Series do begin
   Active               := False;
   SeriesColor          := clBlack;
   LinePen.Style        := psSolid;
   LinePen.Width        := 2;
   LineType             := ltFromPrevious;
   Marks.Visible        := True;
   Pointer.VertSize     := 3;
   Pointer.HorizSize    := 3;
   Pointer.Style        := psCircle;
 end;
end;

procedure TvGraphs.ConfigureBarSeries(var Series: TBarSeries);
begin
  with Series do begin
    Active          := False;
    SeriesColor     := clRed;
    BarPen.Style    := psSolid;
    BarPen.Width    := 1;
    BarWidthPercent := 70;
    BarWidthStyle   := bwPercent;
    Marks.Visible   := True;
  end;
end;

procedure TvGraphs.AssociateTransformations;
begin
  ChtResults.LeftAxis.Transformations          := catResultsLeft;
  ChtResults.AxisList[2].Transformations       := catResultsRight;
  catResultsLeftAutoTransform.Transformations  := catResultsLeft;
  catResultsRightAutoTransform.Transformations := catResultsRight;
end;

procedure TvGraphs.ConfigureTransformations;
begin
     catResultsLeftAutoTransform.MinValue  := 0;
     catResultsLeftAutoTransform.MaxValue  := 1;
     catResultsRightAutoTransform.MinValue := 0;
     catResultsRightAutoTransform.MaxValue := 1;
     catResultsLeftAutoTransform.Enabled   := True;
     catResultsRightAutoTransform.Enabled  := True;
end;

procedure TvGraphs.ConfigureRdgGraphType;
begin
  with rdgGraphType do begin
    Parent       := TabControl;
    Autosize     := False;
    Top          := chtResults.Top + chtResults.Height + 2;
    Left         := 4;
    Height       := 41;
    Width        := 250;
    ColumnLayout := clVerticalThenHorizontal;
    Columns      := 4;
    Visible      := True;
    Items.Add('Line');
    Items.Add('Point');
    Items.Add('Mixed');
    Items.Add('Bar');
    Itemindex    := 0;
    onClick      := @rdgGraphTypeClick;
  end;
end;

procedure TvGraphs.ConfigureChkGridLines;
begin
  with chkGridLinesLeftAxis do begin
    Parent   := TabControl;
    Top      := rdgGraphType.top + rdgGraphType.Height div 3;
    Left     := rdgGraphType.Left + rdgGraphType.Width + 50;
    Height   := 18;
    Width    := 84;
    Checked  := False;
    Caption  := 'Grid Lines (Left)';
    Visible  := True;
    OnChange := @chkGridLinesChange;
  end;

  with chkGridLinesRightAxis do begin
    Parent   := TabControl;
    Top      := rdgGraphType.top + rdgGraphType.Height div 3;
    Left     := chkGridLinesLeftAxis.Left + chkGridLinesLeftAxis.Width + 50;
    Height   := 18;
    Width    := 84;
    Checked  := False;
    Caption  := 'Right';
    Visible  := True;
    OnChange := @chkGridLinesChange;
  end;
end;

procedure TvGraphs.ConfigureChkSeparateGraph;
begin
  with chkSeparateGraph do begin
    Parent   := TabControl;
    Top      := chkGridLinesLeftAxis.top;
    Height   := 18;
    Width    := 75;
    Left     := chtResults.Left + chtResults.Width - chkSeparateGraph.Width - 20;
    Checked  := False;
    Caption  := 'Separate';
    Visible  := True;
    onChange := @chkSeparateGraphChange;
  end;
end;

procedure TvGraphs.CreateYvalsArray;
var
  i,TopYval: integer;
begin
  TopYval := chtResults.Top + 20;
  for i := 0 to MaxFields do begin
    Yvals[i].Xval               := TradioButton.Create(nil);
    Yvals[i].Enabled            := TcheckBox.Create(nil);
    Yvals[i].Colour             := TShape.Create(nil);
    Yvals[i].Xval.Visible       := False;
    Yvals[i].Enabled.Visible    := False;
    Yvals[i].Colour.Visible     := False;
    Yvals[i].Xval.Top           := TopYval + 30*i;
    Yvals[i].Enabled.Top        := TopYval + 30*i;
    Yvals[i].Colour.Top         := TopYval + 30*i;
    Yvals[i].Enabled.Left       := chtResults.Left + chtResults.Width;
    Yvals[i].Colour.Brush.Color := PickColour;
    Yvals[i].Xval.Parent        := tabControl;
    Yvals[i].Enabled.Parent     := tabControl;
    Yvals[i].Colour.Parent      := tabControl;
    Yvals[i].Xval.AutoSize      := True;
    Yvals[i].Enabled.AutoSize   := True;
    Yvals[i].Colour.Height      := Yvals[i].Enabled.Height - 4;
    Yvals[i].Colour.Width       := Yvals[i].Colour.Height;
    Yvals[i].Colour.Left        := Yvals[i].Enabled.Left + 136;
    Yvals[i].Colour.OnClick     := @YvalsColourOnClick;
    Yvals[i].Colour.Hint        := 'Click to change the colour';
    Yvals[i].Colour.ShowHint    := True;
    Yvals[i].Xval.Left          := Yvals[i].Colour.Left + 20;
  end;
end;

procedure TvGraphs.ActivateSeriesForAppropriateAxis(const NumberChecked: integer);
var
  i: integer;
begin
  for i := 0 to MaxFields do begin
    if not Yvals[i].Enabled.Visible then break;
    if Yvals[i].Enabled.Checked then begin
      chtResults.Series.Items[Yvals[i].Series].Active:=True;
      if Yvals[i].Series < 4 then
        ConfigureBarSeries(i, NumberChecked)
      else
        ConfigureLineSeries(i);
    end;
  end;

  AddTopBottomTitles;
  ShowRightGridLinesCheckBox;
end;

procedure TvGraphs.ConfigureBarSeries(const i, NumberChecked: integer);
var
  BarSeries:  TBarSeries;
  BarWidth: integer;
begin
  BarSeries                  := chtResults.Series.Items[Yvals[i].Series] as TBarSeries;
  BarSeries.AxisIndexY       := Yvals[i].Yaxis;
  BarWidth                   := Trunc(100 / NumberChecked);
  BarSeries.BarWidthPercent  := BarWidth;
  BarSeries.BarOffsetPercent := BarWidth*(Yvals[i].Series-Trunc(NumberChecked/2));
  if Trunc(NumberChecked/2)*2 = NumberChecked then
    BarSeries.BarOffsetPercent := BarSeries.BarOffsetPercent + Trunc(BarWidth / 2);
  BarSeries.SeriesColor        := Yvals[i].Colour.Brush.Color;
  chtResults.AxisList.Axes[BarSeries.AxisIndexY].Title.Caption :=
    chtResults.AxisList.Axes[BarSeries.AxisIndexY].Title.Caption + ' '
       + TableData.GetithUFfieldName(Yvals[i].Enabled.Tag);
  chtResults.AxisList.Axes[BarSeries.AxisIndexY].Visible       := True;
  chtResults.AxisList.Axes[BarSeries.AxisIndexY].Title.Visible := True;
end;

procedure TvGraphs.ConfigureLineSeries(const i: integer);
var
  LineSeries: TLineSeries;
begin
  LineSeries             := chtResults.Series.Items[Yvals[i].Series] as TLineSeries;
  LineSeries.AxisIndexY  := Yvals[i].Yaxis;
  LineSeries.SeriesColor := Yvals[i].Colour.Brush.Color;
  LineSeries.Pointer.Brush.Color := Yvals[i].Colour.Brush.Color;
  chtResults.AxisList.Axes[LineSeries.AxisIndexY].Title.Caption :=
    chtResults.AxisList.Axes[LineSeries.AxisIndexY].Title.Caption + ' '
      + TableData.GetithUFfieldName(Yvals[i].Enabled.Tag);
  chtResults.AxisList.Axes[LineSeries.AxisIndexY].Visible       := True;
  chtResults.AxisList.Axes[LineSeries.AxisIndexY].Title.Visible := True;
end;

procedure TvGraphs.ShowRightGridLinesCheckBox;
begin
  chkGridLinesRightAxis.Visible := chtResults.AxisList.Axes[2].Visible;
end;

procedure TvGraphs.YvalsClick(Sender: TObject);
begin
  DrawGraph(False);
end;

procedure TvGraphs.XvalsClick(Sender: TObject);
begin
  CurrPlottedXval := (Sender as TRadioButton).Tag;
  Yvalsclick(Sender);
  end;

procedure TvGraphs.AddTopBottomTitles;
begin
  chtResults.Title.Text.Add(TableData.GetUFtableName);
  chtResults.BottomAxis.Title.Caption := Yvals[CurrPlottedXval].Enabled.Caption;
end;

procedure TvGraphs.AddDataPoints;
var
  i, SeriesCounter: integer;
  j: integer = 4;
  Source:     TListChartSource;
begin
   SeriesCounter := 0;
   for j := 0 to MaxFields do begin
     if not Yvals[j].Enabled.Visible then break;
     if YVals[j].Enabled.Checked then
     begin
       Source := (chtResults.Series.Items[SeriesCounter] as TBarSeries).ListSource;
       chtResults.BottomAxis.Marks.Source := Source;
       CurrPlottedYval := Yvals[j].Enabled.Tag;
       for i := 1 to  StringData.NumberOfRows - 1 do
       begin
         if TableData.GetFieldType(CurrPlottedYval) = vSeconds then
             Source.Add(FormatBottomAxisDataPoint(i), StrToIntDef(FormatSeconds(
               StringData.CellValue(CurrPlottedYval, i)), 0),StringData.CellValue(CurrPlottedXval, i))
         else
           Source.Add(FormatBottomAxisDataPoint(i), StrTointDef(StringData.CellValue(CurrPlottedYval, i), 0),
                StringData.CellValue(CurrPlottedXval, i));
       end;
       SeriesCounter := SeriesCounter + 1;
     end;
   end;
end;

function TvGraphs.FormatBottomAxisDataPoint(const Row: integer): double;
begin
  if TableData.GetFieldType(CurrPlottedXval) = vDate then
    Result := Row
  else
    Result := StrToFloatDef(StringData.CellValue(CurrPlottedXval, Row),0);
end;

procedure TvGraphs.YvalsColourOnClick(Sender: TObject);
begin
  with TColorDialog.Create(nil) do begin
    Color := clYellow;
    if Execute then begin
      (Sender as TShape).Brush.Color:=Color;
      DrawGraph(False);
    end;
  end;
end;

end.

