unit vINI;

{$mode objfpc}{$H+}

////////////////////////////////////////////////////////////////////////
////            Unit providing operations on INI files              ////
////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////
////     Funtions:                                                  ////
////     - Operates on three INI files:                             ////
////       - vfitness.ini                                           ////
////       - filters.ini                                            ////
////       - defaults.ini                                           ////
////                                                                ////
////     vfitness.ini contains info on the tables and the fields    ////
////     filters.ini contains info on the saved view filters        ////
////     defaults.ini contains default values for data entry        ////
////////////////////////////////////////////////////////////////////////


interface

uses
  Classes, stdCtrls, SysUtils, iniFiles, vTypes;

type
  { TvINI }

  TvINI = class
  private
    // Pointer to Vfitness.ini
    INI: TiniFile;
    // Pointer to Defaults.ini
    Dini: TiniFile;
    // Pointer to Filters.ini
    Fini: TiniFile;
    // Open an INI file
    function OpenINIfile(const iniFile: string; CreateNew: boolean = False): TiniFile;
    // Read the table names from vfitness.ini
    function GetTableNames: TstringList;
    // Read the User-friendly table names from vfitness.ini
    function GetUFtableNames: TstringList;
    function GetTableOrder: TstringList;
    procedure SetTableOrder(Avalue: TstringList);
    function GetDefYvalue(const Table: string; const Field: string): boolean;
    function GetDefXvalue(const Table: string; const Field: string): boolean;
    function GetFieldHidden(const Table, Field: string): boolean;
    function GetFieldDefaultValue(const Table, Field: string): string;
    procedure SaveTableFieldData(Table: vTableInfoType);
    procedure SaveFieldData(const Table: string; const NF: integer; const FieldArray: vFieldInfoArrayType);
    procedure SaveAfieldData(const Table: string; const Field: vFieldInfoType);
  public
    constructor Create(const Path:string);
    destructor Destroy; override;
    property TableNames: TstringList read GetTableNames;
    property UFtableNames: TstringList read GetUFtableNames;
    property TableOrder: TstringList read GetTableOrder write SetTableOrder;
    // Read the User-friendly field names from vfitness.ini
    function GetUFfieldNames(const Table: string; const Fields: VfieldArrayType;
      const fieldCount: integer): VfieldArrayType;
    // Return the Field Type (eg String, Description , Seconds)
    function GetFieldType(const Table: string;const  Field: string):vFieldType;
    // Return the Field metadata
    function GetFieldInfo(const Table, Field: string): VfieldInfoType;
    // Get the User-friendly Field name
    function GetUFfieldName(const Table, Field: string): string;
    // Load the names of filters for Table
    procedure LoadFilters(const Table: string; var cmb: TcomboBox);
    // Return the details of a row of a filter
    function AfilterRow(const Row: integer; const Section: string; out filterRow: vFilterRow): boolean;
    procedure SaveFilterRow(const Row: integer; const Section: string; const Table: string; const filterRow: vFilterRow);
    // Return True if a filter with the same name exists
    function FilterExists(const FilterName: string): boolean;
    procedure DeleteFilter(const Section: string);
    procedure SaveFilterOptions(const FilterOptions: TvFilterOptions);
    procedure RestoreFilterOptions(var FilterOptions: TvFilterOptions);
    function IsTableHidden(Table: string): boolean;
    function IsTableGraphed(Table: string): boolean;
    procedure SaveTableInfo(Tables: VTableInfoArrayType);
    function GetFieldOrder(const Table: string): TstringList;
    procedure SetFieldOrder(const Table: string; Avalue: TstringList);
  end;

implementation

{ TvINI }

function TvINI.OpenINIfile(const iniFile: string; CreateNew: boolean = False): TiniFile;
begin
  Result := nil;
  if FileExists(INIfile) or CreateNew then begin
    Result := TiniFile.Create(iniFile);
  end;
end;

constructor TvINI.Create(const Path:string);
begin
  INI  := OpenINIfile(Path + '/vfitness.ini');
  if INI = nil then begin
    raise Exception.Create('vfitness.ini is missing. The programme will close.');
    Halt;
  end;
  Dini := OpenINIfile(Path + '/defaults.ini', True);
{  if Dini = nil then
    raise Exception.Create('Warnng: defaults.ini is missing.');}
  Fini := OpenINIfile(Path + '/filters.ini', True);
{  if Fini = nil then
    raise Exception.Create('filters.ini is missing.');
}
end;

destructor TvINI.Destroy;
begin
  if INI  <> nil then INI.Destroy;
  if Dini <> nil then Dini.Destroy;
  if Fini <> nil then Fini.Destroy;
  inherited Destroy;
end;

function TvINI.GetTableNames: TstringList;
var
  TNames: TstringList;
begin
  TNames := TstringList.Create;
  INI.ReadSection('Tables', TNames);
  Result := TNames;
end;

function TvINI.GetUFtableNames: TstringList;
var
  i: integer;
  TNames, UFtNames: TstringList;
begin
  TNames := TstringList.Create;
  UFtNames := TstringList.Create;
  TNames := GetTableNames;
  for i := 0 to TNames.Count - 1 do
    UFtNames.Add(INI.ReadString('Tables', TNames[i], ''));
  Result := UFtNames;
  Tnames.Destroy;
end;

function TvINI.GetTableOrder: TstringList;
var
  i: integer = 1;
  found: boolean = True;
  Table: string;
  TbleOrder: TstringList;
begin
  TbleOrder := TstringList.Create;
  while found do begin
    Table := INI.ReadString('TableOrder',IntToStr(i),'');
    if Table <> '' then begin
      TbleOrder.Add(Table);
      i := i + 1;
    end
    else
      found := false;
  end;
  if i = 1 then
    TbleOrder := GetTableNames;
  Result := TbleOrder;
end;

procedure TvINI.SetTableOrder(Avalue: TstringList);
var
  i: integer;
begin
  for i := 0 to Avalue.Count-1 do
    INI.WriteString('TableOrder',IntToStr(i+1),Avalue[i]);
end;

function TvINI.GetDefYvalue(const Table: string; const Field: string): boolean;
begin
  Result := CompareText(Field, INI.ReadString('Y', Table, '')) = 0;
end;

function TvINI.GetDefXvalue(const Table: string; const Field: string): boolean;
begin
  Result := CompareText(Field, INI.ReadString('X', Table, '')) = 0;
end;

function TvINI.GetFieldHidden(const Table, Field: string): boolean;
begin
  Result := INI.ReadBool('HiddenFields', Table + '-' + Field, False);
end;

function TvINI.GetFieldDefaultValue(const Table, Field: string): string;
begin
  Result := dINI.ReadString(Table, Field, '');
end;

procedure TvINI.SaveTableFieldData(Table: vTableInfoType);
begin
  if Table.NotGraphed then
    INI.WriteBool('NotGraphedTables', Table.TN, True)
  else
    INI.DeleteKey('NotGraphedTables', Table.TN);
  if Table.Hidden then
    INI.WriteBool('HiddenTables', Table.TN,True)
  else
    INI.DeleteKey('HiddenTables', Table.TN);
  INI.WriteString('Tables', Table.TN, Table.UFN);
  SaveFieldData(Table.TN, Table.NF, Table.FieldInfo);
end;

procedure TvINI.SaveFieldData(const Table: string; const NF: integer; const FieldArray: vFieldInfoArrayType);
var
  i: integer;
begin
  for i := 0 to NF - 1 do
    SaveAfieldData(Table, FieldArray[i]);
end;

procedure TvINI.SaveAfieldData(const Table: string; const Field: vFieldInfoType);
begin
  if Field.FT <> vOther then
    INI.WriteString(SectionNames[Field.FT],Table,Field.FN);
  if Field.Y then
    INI.WriteString('Y',Table,Field.FN);
  if Field.X then
    INI.WriteString('X',Table,Field.FN);
  if Field.UFN <> '' then
    INI.WriteString(Table,Field.FN,Field.UFN)
  else
    INI.DeleteKey(Table,Field.FN);
  if Field.DefaultValue <> '' then
    dINI.WriteString(Table,Field.FN,Field.DefaultValue);
  if Field.Hidden then
    INI.WriteBool('HiddenFields', Table + '-' + Field.FN,True)
  else
    INI.DeleteKey('HiddenFields',  Table + '-' + Field.FN);
end;

function TvINI.GetUFfieldNames(const Table: string; const Fields: VfieldArrayType;
    const fieldCount: integer): VfieldArrayType;
var
  i: integer;
  UFfieldnames: VfieldArrayType;
begin
  for i := 0 to fieldCount-1 do
    UFfieldNames[i].Value := INI.ReadString(Table, Fields[i].Value,Fields[i].Value);
  Result := UFfieldNames;
end;

// Return the type of field (eg string, description, seconds)
function TvINI.GetFieldType(const Table: string; const Field: string): vFieldType;
var
  FieldType: vFieldType;
 Begin
  Result := vOther;
  for FieldType in AllFieldTypes do
    if CompareText(Field, INI.ReadString(SectionNames[FieldType], Table, '')) = 0 then
    begin
      Result := FieldType;
      break;
    end;
end;

function TvINI.GetFieldInfo(const Table, Field: string): VfieldInfoType;
var
  Afield: VfieldInfoType;
begin
  Afield.FN           := Field;
  Afield.UFN          := GetUFfieldName(Table, Field);
  Afield.FT           := GetFieldType(Table, Field);
  Afield.Y            := GetDefYvalue(Table, Field);
  Afield.X            := GetDefXvalue(Table, Field);
  Afield.Hidden       := GetFieldHidden(Table, Field);
  Afield.DefaultValue := GetFieldDefaultValue(Table, Field);
  Result              := Afield;
end;

function TvINI.GetUFfieldName(const Table, Field: string): string;
begin
  Result := INI.ReadString(Table, Field, Field);
end;

procedure TvINI.LoadFilters(const Table: string; var cmb: TcomboBox);
var
  i: integer;
  Sections: TStringList;
begin
  cmb.Clear;
  Sections := TStringList.Create;
  Fini.ReadSections(Sections);
  for i := 0 to Sections.Count - 1 do
  begin
    if CompareText(Fini.ReadString(Sections.Strings[i], 'table', ''),
      Table) = 0 then
      cmb.Items.Add(Sections.Strings[i]);
  end;
  Sections.Destroy;
end;

function TvINI.AfilterRow(const Row: integer; const Section: string; out filterRow: vFilterRow): boolean;
begin
  Result := False;
  filterRow.Quantity   := Fini.ReadString(Section,  'Quantity' + IntToStr(Row), '');
  filterRow.Field      := Fini.ReadString(Section, 'Field' + IntToStr(Row), '');
  filterRow.Comparison := Fini.ReadString(Section, 'Comparison' + IntToStr(Row), '');
  filterRow.Course     := Fini.ReadInteger(Section, 'Course', -1);
  filterRow.DateFrom   := Fini.ReadString(Section,  'DateFrom', '');
  filterRow.DateTo     := Fini.ReadString(Section,  'DateTo', '');
  if FilterRow.Quantity = '' then exit;
  Result := True;
end;

procedure TvINI.SaveFilterRow(const Row: integer; const Section: string; const Table: string; const filterRow: vFilterRow);
begin
  Fini.WriteString(Section,  'Quantity' + IntToStr(Row),   FilterRow.Quantity);
  Fini.WriteString(Section, 'Field' + IntToStr(Row),      FilterRow.Field);
  Fini.WriteString(Section, 'Comparison' + IntToStr(Row), FilterRow.Comparison);
  if FilterRow.Course <> -1 then
    Fini.WriteInteger(Section, 'Course',                     FilterRow.Course);
  if FilterRow.DateFrom <> '' then
    Fini.WriteString(Section,  'DateFrom',                   FilterRow.DateFrom);
  if FilterRow.DateTo <> '' then
    Fini.WriteString(Section,  'DateTo',                     FilterRow.DateTo);
  Fini.WriteString(Section, 'table',                       Table);
end;

function TvINI.FilterExists(const FilterName: string): boolean;
begin
  Result := FIni.ReadString(FilterName, 'table', '') <> '';
end;

procedure TvINI.DeleteFilter(const Section: string);
begin
  Fini.EraseSection(Section);
end;

procedure TvINI.SaveFilterOptions(const FilterOptions: TvFilterOptions);
begin
  Fini.WriteBool('Options', 'Course',FilterOptions.StickyCourse);
  Fini.WriteBool('Options', 'Dates',FilterOptions.StickyDates);
end;

procedure TvINI.RestoreFilterOptions(var FilterOptions: TvFilterOptions);
begin
  FilterOptions.StickyCourse := Fini.ReadBool('Options', 'Course', True);
  FilterOptions.StickyDates  := Fini.ReadBool('Options', 'Dates', True);
end;

function TvINI.IsTableHidden(Table: string): boolean;
begin
  Result := INI.ReadBool('HiddenTables', Table, False);
end;

function TvINI.IsTableGraphed(Table: string): boolean;
begin
  Result := not INI.ReadBool('NotGraphedTables', Table, False);
end;

procedure TvINI.SaveTableInfo(Tables: VTableInfoArrayType);
var
  i: integer;
begin
  for i := 0 to High(Tables) do
    if Tables[i].NF = 0 then
      break
    else
      SaveTableFieldData(Tables[i]);
end;

function TvINI.GetFieldOrder(const Table: string): TstringList;
var
  i: integer = 1;
  found: boolean = True;
  Fld: string;
  FldOrder: TstringList;
begin
  FldOrder := TstringList.Create;
  while found do begin
    Fld := INI.ReadString(Table+'-FieldOrder',IntToStr(i),'');
    if Fld <> '' then begin
      FldOrder.Add(Fld);
      i := i + 1;
    end
    else
      found := false;
  end;
  Result := FldOrder;
end;

procedure TvINI.SetFieldOrder(const Table: string; Avalue: TstringList);
var
  i: integer;
begin
  for i := 0 to Avalue.Count-1 do
    INI.WriteString(Table + '-FieldOrder',IntToStr(i+1),Avalue[i]);
end;

end.

