unit vsqlite3;

{$mode objfpc} {$H+}

///////////////////////////////////////////////////////////////
////       Unit to provide access to the physical DB       ////
///////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////
////     Functions:                                        ////
////     - Open the database...looks for a 'config' table  ////
////     - Open a table                                    ////
////     - load the data from a table into a StringGrid    ////
////     - Open the 'Courses' table...into an array        ////
////     - provide info such as fieldcount, recordcount    ////
////                                                       ////
////     The unit by default runs in safe mode, that is,   ////
////        the database is kept closed between operations ////
///////////////////////////////////////////////////////////////

interface

uses
  Classes, StrUtils, SysUtils, Grids, Sqlite3DS, Dialogs, vTypes;

type

  { Tvsqlite3 }

{
 Operation:
   - call Create to create the class, specifying the full path to an sqlite3 database
   - call OpenTable to open a particular table in the database
   - call LoadTable to load the data into a StringGrid
   - call LoadCourseTable to load the Courses Table into an Array
   - use the table to save data, delete data, etc
   - call Destroy to tidy up.
}

  Tvsqlite3 = class
  private
    cw: string;
    // Pointer to the sqlite database
    dst: TSqlite3Dataset;
    // Field Count
    fc: integer;
    // Record Count
    rc: integer;
    // Current Table
    ct: string;
    // Safe Mode, that is, True if the Database is closed when not being used.
    sf: boolean;
    // The names of the Fields in CurrentTable. NB Fields[0] = the ID, which is unique.
    Fields: vFieldArrayType;
    procedure SetCurrentTable(const AValue: string);
    procedure SetCurrentWhere(const AValue: string);
    procedure SetFieldCount(const AValue: integer);
    procedure SetRecordCount(const AValue: integer);
    procedure SetSafeMode(const AValue: boolean);
    // Save field names in array
    procedure SaveFieldNames;
    // Return the name of the i-th Field (zero-based index)
    function GetFieldName(const i: integer): string;
    // Save changes by running the SQL string
    function SaveData(const SQL: string): boolean;
    // If SaveMode is true, makes sure that the table is closed after an operation
    procedure CheckSafeMode(const OpenIfClosed: boolean = False);
    // Load the headers into the first row of a stringGrid
    procedure LoadHeaders(sgd: TstringGrid);
    // Get the name of the ith Field, as stored in 'Fields;
    function GetStoredFieldName(const i: integer): string;
    // load data from table into a string grid
    procedure LoadTable(sgd: TstringGrid);
    // if data contains quotes, add another quote for SQL compatability
    function AddAnotherQuote(const FieldValue: string): string;
    // Open the View using the table 'TableName' and the 'Where' clause.
    // Returns the Record Count.
    function OpenTable(out FeeldCount: integer; const TableName: string; const OrderBy: string = '';
      Where: string = ''): integer;
    // Return the Value of the ith Field of the table, given the fields as in the TableData records
    function ithfieldValue(const Feelds: vFieldArrayType; const FieldNdx: integer): string;
  public
    // The 'Create' constructor. 'KeepClosed' is True means that the
    // table is closed in between operations.
    constructor Create(const DBname: string; KeepClosed: boolean = True);
    destructor Destroy; override;
    property FieldCount: integer read fc write SetFieldCount;
    property RecordCount: integer read rc write SetRecordCount;
    // SafeMode = True means that the table is closed between operations
    property SafeMode: boolean read sf write SetSafeMode;
    property CurrentTable: string read ct write SetCurrentTable;
    property CurrentWhere: string read cw write SetCurrentWhere;
    // Save the data in FieldValues to the CurrentTable. If the first field (ID)
    // is empty, then do an insert, otherwise do an update.
    function SaveData(const FieldValues: VFieldArrayType): boolean;
    function SaveData(const FieldValues: VFieldArrayType; Table: string): boolean;
    // Delete the record of the CurrentTable specified by the ID (first field)
    procedure DeleteData(const ID: string);
    // load data from table Table into a string grid
    procedure LoadTable(sgd: TstringGrid; const Table: string; const OrderBy: string = ''; const Where: string = '');
    // Load data from the Courses table into the 'Courses' array
    function LoadCourseTable(var Courses: VCourseArrayType): integer;
    // Return the Field names
    function FieldNames: VfieldArrayType;
    function FieldNames(out FeeldCount: integer; Table: string): VfieldArrayType;
  end;

{var
  vvSqlite3: Tvsqlite3;
}
implementation

constructor Tvsqlite3.Create(const DBname: string; KeepClosed: boolean = True);
begin
  if DBname = '' then
    raise Exception.Create('''DBname'' may not be null in Tvsqlite3.create');
  dst := tsqlite3dataset.Create(nil);
  try
    dst.FileName := DBname;
    dst.TableName := 'Config';
    dst.Open;
    CurrentTable := '';
    CurrentWhere := '';
    RecordCount := 0;
    FieldCount := 0;
  except
    raise Exception.Create('''' + DBname + '''' +
      ' is not a valid SQLite3 Database (Tvsqlite3.create).' + LineEnding +
      'Check that the ' + '''' + 'Config' + '''' + ' table exists');
  end;

  SafeMode := KeepClosed;
  CheckSafeMode(False);
end;

destructor Tvsqlite3.Destroy;
begin
  dst.Close;
  dst.Destroy;
  inherited Destroy;
end;

procedure Tvsqlite3.CheckSafeMode(const OpenIfClosed: boolean = False);
begin
  if SafeMode then
    dst.Active := OpenIfClosed
  else
    dst.Active := True;
end;

procedure Tvsqlite3.LoadHeaders(sgd: TstringGrid);
var
  i: integer;
begin
  for i := 0 to sgd.ColCount - 1 do
    Sgd.cells[i, 0] := GetStoredFieldName(i);
end;

function Tvsqlite3.GetStoredFieldName(const i: integer): string;
begin
  Result := Fields[i].Value;
end;

function Tvsqlite3.SaveData(const SQL: string): boolean;
begin
  Result := True;
  CheckSafeMode(True);
  try
    dst.ExecSQL(SQL);
    dst.ApplyUpdates;
  except
    Result := False;
  end;
  CheckSafeMode(False);
end;

function Tvsqlite3.SaveData(const FieldValues: VFieldArrayType): boolean;
var
  i: integer;
  SQL: string;
begin
  if FieldValues[0].Value = '' then
  begin // Insert record
    SQL := 'insert into ' + CurrentTable + ' ( ';
    for i := 1 to FieldCount - 1 do
      SQL := SQL + ifthen(i <> 1, ', ''', '''') + Fields[i].Value + ''' ';
    SQL := SQL + ') values (';
    for i := 1 to FieldCount - 1 do
    begin
      SQL := SQL + ifthen(i <> 1, ', ''', ' ''') + AddAnotherQuote(ithfieldValue(FieldValues,i)) + ''' ';
    end;
    SQL := SQL + ');';
  end
  else
  begin // Update record
    SQL := 'update ' + CurrentTable + ' SET ';
    for i := 1 to FieldCount - 1 do
    begin
      SQL := SQL + ifthen(i <> 1, ', ''', '''') + Fields[i].Value +
        ''' = ''' + AddAnotherQuote(ithfieldValue(FieldValues,i)) + ''' ';
    end;
    SQL := SQL + ' where ' + Fields[0].Value + ' = ''' + FieldValues[0].Value + ''';';
  end;
  Result := SaveData(SQL);
end;

function Tvsqlite3.SaveData(const FieldValues: VFieldArrayType; Table: string
  ): boolean;
var
  Ttable: string;
begin
  Ttable := CurrentTable;
  OpenTable(fc, Table);
  SaveData(FieldValues);
  OpenTable(fc, Ttable, CurrentWhere);
  Result := True;
end;

procedure Tvsqlite3.DeleteData(const ID: string);
begin
  SaveData('delete from ' + CurrentTable + ' where ' + Fields[0].Value +
    ' = ''' + ID + ''';');
end;

function Tvsqlite3.OpenTable(out FeeldCount: integer; const TableName: string; const OrderBy: string = '';
  Where: string = ''): integer;
begin
  dst.active := False;
  Where := ifthen(Where = '',Where,' where ' + where);
  Where := ifthen(OrderBy = '',Where,Where + ' order by ' + OrderBy);
  dst.SQL := 'select * from ' + TableName + Where;
  dst.TableName := TableName;
  dst.Active := True;
  FieldCount := dst.FieldCount;

  if FieldCount > MaxFields then
    raise Exception.Create('Maximum number of fields reached for ' + CurrentTable);

  Feeldcount := FieldCount;
  RecordCount := dst.RecordCount;
  CurrentTable := TableName;
  CurrentWhere := Where;
  SaveFieldNames;
  CheckSafeMode(False);
  Result := RecordCount;
end;

function Tvsqlite3.ithfieldValue(const Feelds: vFieldArrayType; const FieldNdx: integer): string;
var
  i: integer = 0;
begin
  while (i < FieldCount) and
        (CompareText(Feelds[i].Name,Fields[FieldNdx].Value) <> 0) do
    i := i + 1;
  Result := Feelds[i].Value;
end;

function Tvsqlite3.GetFieldName(const i: integer): string;
begin
  Result := dst.FieldDefs.Items[i].DisplayName;
end;

procedure Tvsqlite3.LoadTable(sgd: TstringGrid);
var
  arow: integer = 1;
  i: integer;
begin
  CheckSafeMode(True);
  sgd.ColCount := FieldCount;
  // Add an extra row for the header/column titles
  sgd.RowCount := RecordCount + 1;
  LoadHeaders(sgd);
  dst.First;
  while not dst.EOF do
  begin
    for i := 0 to sgd.ColCount - 1 do
      Sgd.cells[i, arow] := dst.Fields[i].AsString;
    dst.Next;
    arow := arow + 1;
  end;
  CheckSafeMode(False);
end;

function Tvsqlite3.AddAnotherQuote(const FieldValue: string):string;
var
  j, k: integer;
  ts : string;
begin
    // Add another quote if one found
    ts := FieldValue;
    k := 1;
    j := 0;
    repeat
      j := Npos('''', ts, k);
      if j > 0 then
        ts := StuffString(ts, j, 1, '''''');
      k := k + 2;
    until j <= 0;
    Result := ts;
end;

procedure Tvsqlite3.LoadTable(sgd: TstringGrid; const Table: string; const OrderBy: string = ''; const Where: string = '');
begin
  RecordCount := OpenTable(fc, Table, OrderBy, Where);
  LoadTable(sgd);
end;

procedure Tvsqlite3.SetFieldCount(const AValue: integer);
begin
  fc := AValue;
end;

procedure Tvsqlite3.SetCurrentTable(const AValue: string);
begin
  ct := AValue;
end;

procedure Tvsqlite3.SetCurrentWhere(const AValue: string);
begin
  cw:=AValue;
end;

procedure Tvsqlite3.SetRecordCount(const AValue: integer);
begin
  rc := AValue;
end;

procedure Tvsqlite3.SetSafeMode(const AValue: boolean);
begin
  sf := AValue;
end;

procedure Tvsqlite3.SaveFieldNames;
var
  i: integer;
begin
  CheckSafeMode(True);
  for i := 0 to FieldCount - 1 do
    Fields[i].Value := GetFieldName(i);
  CheckSafeMode(False);
end;

function Tvsqlite3.LoadCourseTable(var Courses: VCourseArrayType): integer;
var
  dstT: Tsqlite3Dataset = nil;
  i: integer = 0;
begin
  if compareText(dst.TableName, 'Courses') <> 0 then
  begin
    dstT := tsqlite3dataset.Create(nil);
    dstT.TableName := dst.Tablename;
    dstT.SQL := dst.SQL;
  end;

  dst.Close;
  dst.TableName := 'Courses';
  dst.SQL := 'select * from courses';
  dst.Open;

  // Check that there are courses to display
  Result := 0;
  if dst.RecordCount = 0 then
    exit;

  // Read the course info into Courses
  dst.First;
  while not dst.EOF do
  begin
    Courses[i].Description := dst.FieldByName('Description').AsString;
    Courses[i].Name := dst.FieldByName('CourseName').AsString;
    Courses[i].ID := dst.FieldByName('Coursesid').AsString;
    Courses[i].Distance := dst.FieldByName('Distance').AsString;
    Courses[i].URL := dst.FieldByName('URL').AsString;
    i := i + 1;
    if i > MaxCourses then break;
    dst.Next;
  end;
  Result := i;

  if dstT <> nil then
  begin
    dst.Close;
    dst.TableName := dstT.TableName;
    dst.SQL := dstT.SQL;
    CheckSafeMode(False);
    FreeAndNil(dstT);
  end;
end;

function Tvsqlite3.FieldNames: VfieldArrayType;
begin
  Result := Fields;
end;

function Tvsqlite3.FieldNames(out FeeldCount: integer; Table: string): VfieldArrayType;
begin
  OpenTable(FeeldCount, Table);
  Result := Fields;
end;

end.

