unit vstringdata; 

{$mode objfpc}  {$H+}

interface

////////////////////////////////////////////////////////////////////////////
// Unit providing operations on the loaded data as stored in a StringGrid //
////////////////////////////////////////////////////////////////////////////

uses
  Classes, SysUtils, dialogs, Grids, vTypes, StdCtrls, vSqlite3, vTableData, vCourses, vUtils, graphics;

type

  { TvStringData }

  // Function definition for 'formatdate' and 'formatseconds' used to configure date and seconds columns
  TvFormat = function(item: string; forDisplay: boolean = False): string;

  TvStringData = class
  private
    // Used by the private CurrentWhere property
    FCurrentWhere: string;
    // The two string grids, containing the data, and the stats (eg average time for all records)
    StringGrid:    TstringGrid;
    StatsGrid:     TstringGrid;
    btnMap:        Tbutton;
    // Variables for the classes in other units
    sqlite3:       TvSqlite3;
    courseData:    TvCourses;
    TableData:     TvTableData;
    // The column number of the last Sorted Column
    SortedColumn:  integer;
    // Used by the private CurrentTable property
    FTableName: string;
    // The stringGrid column 'i' displays the data for the field ColumnsFields[i]
    // -1 is the value of the element after the last column.
    ColumnsFields: array[0..MaxFields+1] of integer;
    // Callback function type, provided by vGraphs in the RegisterDrawGraph function
    DrawGraph: TvDrawGraph;
    // Used by the public property GetFields
    function GetFields: vFieldArrayType;
    // Initialize a vFieldArrayType variable to empty strings
    procedure InitvFieldArrayType(out Fields: vFieldArrayType);
    // Used by the public property NumberOfRows
    function GetNumberOfRows: integer;
    // Used by the private CurrentTable property
    procedure SetTableName(const AValue: string);
    // Used by the private CurrentWhere property
    procedure SetCurrentWhere(const AValue: string);

    // Properties
    property CurrentTable: string read FTableName write SetTableName;
    property CurrentWhere: string read FCurrentWhere write SetCurrentWhere;

    // Change the headers to use the User-friendly field names
    procedure AddUFfieldNames;
    // Hide the first (ID) column
    procedure HideIDcolumn;
    // Hide the columns whose fields are marked as hidden
    procedure HideOtherColumns;
    // Configure a column for user-friendly display (eg change seconds to minutes:seconds)
    procedure ConfigureColumn(Format: TvFormat; const FieldType: vFieldType);
    // Configure a field (about to be saved) for database format (eg change minutes:seconds to seconds)
    procedure ConfigureField(var Fields: vFieldArrayType; Table: string; Format: TvFormat; const FieldType: vFieldType);
    // Make the dates in the 'date' column more user-friendly
    procedure ConfigureDateColumn;
    // Make the seconds in the 'seconds' column display as minutes:seconds
    procedure ConfigureSecsColumn;
    // Convert the dates in the 'date' field to database format (yyyymmdd)
    procedure ConfigureDateFIeld(var Fields: vFieldArrayType; Table: string = '');
    // Convert the seconds in the 'seconds' field to database format (total seconds)
    procedure ConfigureSecsField(var Fields: vFieldArrayType; Table: string = '');
    // Find which column is of fieldtype 'vfieldType' eg vDate for a Date column
    function FindColumn(const FieldType: vFieldType; Table: string = ''): integer;
    // Change the Course ID to a User-friendly Course name for each record
    procedure ConvertCourseIDtoName;
    // Narrow the width of URL columns
    procedure ShortenURLcolumn;
    // Lengthen the Description (last) column to fill the String Grid
    procedure LengthenDescriptionColumn;
    // Configure the two grids to synchronize their movements and column widths
    procedure SyncStatsGrid;
    // Put the current cell in the String Grid to last row and first displayed column
    procedure PositionCursor;
    // Configure the Grids (eg point functions to string grid events)
    procedure ConfigureGrids;
    // procedure for onTopLeftChanged for String Grid - ensure that the Stats
    // grid scrolls with the String Grid
    procedure StringGridTopLeftChanged(Sender:Tobject);
    // algorithms for comparing cells when sorting after clicking header
    procedure StringGridCompareCells(Sender: TObject; ACol, ARow, BCol,
      BRow: Integer; var Result: integer);
    // Sort the string grid on the column clicked on
    procedure StringGridHeaderClick(Sender: TObject; IsColumn: Boolean;
      Index: Integer);
    // Calculate the Statistics for the current table
    procedure CalcStats;
    // Save the stats for a cell of String Grid
    procedure SaveStats(const FieldIndex: integer; const Value: string);
    // Display the Statistics Labels (eg Count, Maximum, Average)
    procedure DisplayStatsLabels;
    // Clear a column in the Stats Grid
    procedure ClearStatsColumn(const Column: integer);
    // Display a column in the Stats Grid
    procedure DisplayStatsColumn(const Column: integer; const SecondsField: boolean);
    // Turns map button on/off when stringgrid row changes
    procedure ResultsGridSelection(Sender: TObject; aCol, aRow: integer);
    // Check current selection for existence of map of course
    procedure ShowBtnMap(const Row: integer);
    // Open URL for map
    procedure btnMapClick(Sender: TObject);
    // Convert Field index to the Column index, and vice versa
    function FieldNdx2ColNdx(const FieldIndex: integer): integer;
    function ColNdx2FieldNdx(const ColumnIndex: integer): integer;
    // Delete a column from the StringGrid (for when the field is marked as hidden)
    procedure DeleteColumn(col: integer);
    // Initialize ColumnsFields so that column i points to field i
    procedure InitColumnsFields;
    // Move Columns to match Field Order
    procedure AlignColumnsWithFields;
    // Load the current table into the String Grid
    procedure LoadTable;
    // Colour the header of the sorted column in Ligh Olive Green
    procedure StringGridPrepareCanvas(sender: TObject; aCol, aRow: Integer; aState: TGridDrawState);
  public
    constructor Create(StringGridControl, StatsGridControl:TstringGrid; MapButton: Tbutton; sqlite3DB:Tvsqlite3;
      TData:TvTableData; Cdata: TvCourses);
    destructor Destroy; override;
    // Return the Values of the current row of the String Grid
    // To be fixed, so as to also pass the field for each column.
    property OneRecord: vFieldArrayType read GetFields;
    // Return the number of Rows in the String Grid
    property NumberOfRows: integer read GetNumberOfRows;
    // Load the table 'Table' into the String Grid
    procedure LoadTable(const Table: string; const Where: string = '');
    // Navigate through records of the StringGrid
    procedure GoFirst;
    procedure GoLast;
    procedure GoNext;
    procedure GoPrevious;
    // All data is saved in this routine. Fix this, when fixing addeditGUI for editing only some fields.
    procedure SaveData(Fields: vFieldArrayType; Table: string = '');
    // Delete a row of StringGrid
    procedure DeleteData;
    // The value of a cell, used by vGraph. Needs checking to see if it takes into
    // account that the column and the field may not be the same index.
    function CellValue(const FieldNdx, row: integer): string;
    // The DrawGraph function is called when a column is sorted.
    procedure RegisterDrawGraph(DrawGrph: TvDrawGraph);
  end;

implementation

uses StrUtils;

{ TvStringData }

// Set the current table name in this unit as well as the vTableData unit
procedure TvStringData.SetTableName(const AValue: string);
begin
  FTableName:=AValue;
end;

procedure TvStringData.SetCurrentWhere(const AValue: string);
begin
  FCurrentWhere:=AValue;
end;

function TvStringData.GetFields: vFieldArrayType;
var
  Fields: vFieldArrayType;
  i:integer;
begin
  InitvFieldArrayType(Fields);
  for i := 0 to StringGrid.ColCount -1 do
    Fields[ColNdx2FieldNdx(i)].Value := StringGrid.Cells[i,StringGrid.Row];
  Result := Fields;
end;

procedure TvStringData.InitvFieldArrayType(out Fields: vFieldArrayType);
var
  i: integer;
begin
  for i := 0 to MaxFields do Fields[i].Value := '';
end;

function TvStringData.GetNumberOfRows: integer;
begin
  Result := StringGrid.RowCount;
end;

procedure TvStringData.AddUFfieldNames;
var
  i:integer;
  Fields: VfieldArrayType;
begin
  Fields := TableData.GetUFfieldNames;
  for i := 0 to StringGrid.ColCount - 1 do
    StringGrid.Cells[i, 0] := Fields[ColNdx2FieldNdx(i)].Value;
end;

procedure TvStringData.HideIDcolumn;
begin
  StringGrid.ColWidths[0] := 0;
end;

procedure TvStringData.HideOtherColumns;
var
  i: integer = 1;
begin
  while i < StringGrid.ColCount do
    if TableData.GetFieldHidden(ColNdx2FieldNdx(i)) then
      DeleteColumn(i)
    else
      i := i + 1;
end;

procedure TvStringData.ConfigureColumn(Format: TvFormat; const FieldType: vFieldType);
var
  Col, i : integer;
begin
  Col := FindColumn(FieldType);
  if Col = -1 then exit;
  for i := 1 to StringGrid.rowCount - 1 do
    StringGrid.Cells[Col,i] := Format(StringGrid.Cells[Col,i],True);
end;

procedure TvStringData.ConfigureField(var Fields: vFieldArrayType; Table: string;
  Format: TvFormat; const FieldType: vFieldType);
var
  Ndx : integer;
begin
  Ndx := FindColumn(FieldType, Table);
  if Ndx = -1 then exit;
  Ndx := ColNdx2FieldNdx(Ndx);
  Fields[Ndx].Value := Format(Fields[Ndx].Value,False);
end;

procedure TvStringData.ConfigureDateColumn;
begin
  ConfigureColumn(@FormatDate,vDate);
end;

procedure TvStringData.ConfigureSecsColumn;
begin
  ConfigureColumn(@FormatSeconds,vSeconds);
end;

procedure TvStringData.ConfigureDateFIeld(var Fields: vFieldArrayType; Table: string = '');
begin
  ConfigureField(Fields, Table, @FormatDate,vDate);
end;

procedure TvStringData.ConfigureSecsField(var Fields: vFieldArrayType; Table: string = '');
begin
  ConfigureField(Fields, Table, @FormatSeconds,vSeconds);
end;

function TvStringData.FindColumn(const FieldType: vFieldType; Table: string = ''): integer;
begin
  Result := FieldNdx2ColNdx(TableData.IndexOfField(FieldType, Table));
end;

constructor TvStringData.Create(StringGridControl,StatsGridControl: TstringGrid; MapButton: Tbutton;
  sqlite3DB: Tvsqlite3;TData:TvTableData; Cdata: TvCourses);
begin
  StringGrid     := StringGridControl;
  StatsGrid      := StatsGridControl;
  btnMap         := MapButton;
  btnMap.OnClick := @btnMapClick;
  sqlite3        := sqlite3DB;
  CourseData     := Cdata;
  TableData      := TData;
  CurrentTable   := TableData.FirstTable;
  DrawGraph      := nil;
  ConfigureGrids;
  LoadTable;
end;

destructor TvStringData.Destroy;
begin
  inherited Destroy;
end;

procedure TvStringData.LoadTable(const Table: string; const Where: string = '');
begin
  CurrentTable  := ifthen(Table = '', CurrentTable, Table);
  CurrentWhere := Where;
  LoadTable;
end;

procedure TvStringData.GoFirst;
begin
  StringGrid.Row := 1;
end;

procedure TvStringData.GoLast;
begin
  StringGrid.Row := StringGrid.RowCount - 1;
end;

procedure TvStringData.GoNext;
begin
  StringGrid.Row := StringGrid.Row + 1;
end;

procedure TvStringData.GoPrevious;
begin
  StringGrid.Row := StringGrid.Row - 1;
end;

// This needs to be improved to take into account which fields are being saved.
procedure TvStringData.SaveData(Fields: vFieldArrayType; Table: string = '');
var
  TableToUse: string;
begin
  TableData.AddFieldNames(Fields, Table);
  ConfigureDateField(Fields, Table);
  ConfigureSecsField(Fields, Table);
  TableToUse := ifthen(Table = '',CurrentTable,Table);
  if compareText(CurrentTable,TableToUse) = 0 then begin
    sqlite3.SaveData(Fields);
    Loadtable;
  end
  else
    sqlite3.SaveData(Fields, TableToUse);
  // If the Courses table has had it's data changed, reload all of the Courses ComboBoxes.
  if CompareText(TableToUse,CourseData.TableName) = 0 then
    CourseData.Reload;
end;

procedure TvStringData.DeleteData;
begin
  sqlite3.DeleteData(StringGrid.cells[0, StringGrid.Row]);
  LoadTable;
  // If the Courses table has had it's data changed, reload all of the Courses ComboBoxes.
  if CompareText(CurrentTable,CourseData.TableName) = 0 then
    CourseData.Reload;
end;

function TvStringData.CellValue(const FieldNdx, row: integer): string;
begin
  Result :=  StringGrid.Cells[FieldNdx2ColNdx(FieldNdx), row];
end;

procedure TvStringData.RegisterDrawGraph(DrawGrph: TvDrawGraph);
begin
  DrawGraph := DrawGrph;
end;

procedure TvStringData.LoadTable;
begin
  SortedColumn   := 1;
  sqlite3.LoadTable(StringGrid, CurrentTable, TableData.IthFieldName(1), CurrentWhere);
  InitColumnsFields;
  AlignColumnsWithFields;
  AddUFfieldNames;
  ConfigureDateColumn;
  ConfigureSecsColumn;
  ConvertCourseIDtoName;
  PositionCursor; // NB Put this before any col width changes.
  ShortenURLcolumn;
  LengthenDescriptionColumn;
  HideIDcolumn;
  HideOtherColumns;
  CalcStats;
  SyncStatsGrid;    // Put this last
  ShowBtnMap(StringGrid.RowCount-1);
end;

procedure TvStringData.StringGridPrepareCanvas(sender: TObject; aCol,
  aRow: Integer; aState: TGridDrawState);
begin
  If ((aCol = SortedColumn) and (aRow = 0)) then
      StringGrid.Canvas.Brush.Color := clMoneyGreen;
  Unused(Sender, aState);
end;

procedure TvStringData.ConvertCourseIDtoName;
var
  CourseCol, i:integer;
begin
  CourseCol := FindColumn(vCourse);
  if CourseCol = -1 then exit;
  for i := 1 to StringGrid.RowCount -1 do
    StringGrid.Cells[CourseCol,i] := CourseData.CourseName(StrToInt(StringGrid.Cells[CourseCol,i]));
end;

procedure TvStringData.ShortenURLcolumn;
var
  URLindex: integer;
begin
  URLindex := TableData.IndexOfField(vURL);
  if URLindex = -1 then exit;
  StringGrid.ColWidths[FieldNdx2ColNdx(URLindex)] := 100;
end;

procedure TvStringData.LengthenDescriptionColumn;
begin
  if TableData.GetFieldType(ColNdx2FieldNdx(StringGrid.ColCount - 1)) <> vDescription then exit;

  if StringGrid.CellRect(StringGrid.ColCount - 1, 0).Right <
    (StringGrid.Left + StringGrid.Width) then
      StringGrid.ColWidths[StringGrid.ColCount - 1] :=
        StringGrid.Width + StringGrid.Left - StringGrid.CellRect(
        StringGrid.ColCount - 2, 0).Right;
end;

procedure TvStringData.SyncStatsGrid;
var
  i: integer;
begin
  StatsGrid.AutoSizeColumns;
  StatsGrid.ColWidths[0] := 0;
  StatsGrid.ColCount     := StringGrid.ColCount;
  StatsGrid.Width        := StringGrid.Width;
  for i := 1 to StringGrid.ColCount -1 do
    if StatsGrid.ColWidths[i] < StringGrid.ColWidths[i] then
      StatsGrid.ColWidths[i] := StringGrid.ColWidths[i]
    else
      StringGrid.ColWidths[i] := StatsGrid.ColWidths[i]
end;

procedure TvStringData.PositionCursor;
begin
  StringGrid.AutoSizeColumns;
  StringGrid.Col := 1;
  StringGrid.Row := Pred(StringGrid.RowCount);
end;

procedure TvStringData.ConfigureGrids;
begin
  StringGrid.OnTopLeftChanged := @StringGridTopLeftChanged;
  StringGrid.OnHeaderClick    := @StringGridHeaderClick;
  StringGrid.OnCompareCells   := @StringGridCompareCells;
  StatsGrid.RowCount          := 5;
  StatsGrid.Cells[1, 0]       := 'Count';
  StatsGrid.Cells[1, 1]       := 'Average';
  StatsGrid.Cells[1, 2]       := 'Total';
  StatsGrid.Cells[1, 3]       := 'Minimum';
  StatsGrid.Cells[1, 4]       := 'Maximum';
  StringGrid.OnSelection      := @ResultsGridSelection;
  StringGrid.OnPrepareCanvas  := @StringGridPrepareCanvas;
end;

procedure TvStringData.StringGridTopLeftChanged(Sender: TObject);
begin
  StatsGrid.LeftCol := StringGrid.LeftCol;
end;

procedure TvStringData.StringGridCompareCells(Sender: TObject; ACol, ARow, BCol,
  BRow: Integer; var Result: integer);
begin
  case TableData.GetFieldType(ColNdx2FieldNdx(Acol)) of
    vString, vDescription, vCourse, vURL: Result :=
        CompareText(StringGrid.cells[ACol, Arow], StringGrid.cells[BCol, Brow]);
    vDate: Result :=
        Round(StrToFloatDef(FormatDate(StringGrid.Cells[ACol, ARow]), 0) -
        StrToFloatDef(FormatDate(StringGrid.Cells[BCol, BRow]), 0));
    vSeconds: Result :=
        Round(StrToFloatDef(FormatSeconds(StringGrid.Cells[ACol, ARow]), 0) -
        StrToFloatDef(FormatSeconds(StringGrid.Cells[BCol, BRow]), 0));
    else
      Result :=
        Round(StrToFloatDef(StringGrid.Cells[ACol, ARow], 0) -
        StrToFloatDef(StringGrid.Cells[BCol, BRow], 0));
  end;
  if StringGrid.SortOrder = soDescending then
    Result := -Result;
end;

procedure TvStringData.StringGridHeaderClick(Sender: TObject; IsColumn: Boolean;
  Index: Integer);
begin
  if (StringGrid.SortOrder = soDescending) or (SortedColumn <> Index) then
  begin
    StringGrid.SortOrder := soAscending;
    SortedColumn := Index;
  end
  else
    StringGrid.SortOrder := soDescending;
  StringGrid.SortColRow(True, Index, StringGrid.FixedRows, StringGrid.RowCount - 1);
  if DrawGraph <> nil then DrawGraph(False);
  Unused(IsColumn);
end;

procedure TvStringData.CalcStats;
var
  i,j, FieldNdx: integer;
begin
  StatsGrid.ColCount := StringGrid.ColCount;
  TableData.ZeroStatsTotals;
  for j := 1 to StringGrid.RowCount - 1 do
    for i := 0 to StringGrid.ColCount - 1 do
    begin
      FieldNdx := ColNdx2FieldNdx(i);
      if StringGrid.Cells[i,j] <> '' then
          TableData.IncrementNZC(FieldNdx);
      if TableData.GetFieldType(FieldNdx) = vSeconds then
        SaveStats(FieldNdx,FormatSeconds(StringGrid.Cells[i,j]))
      else
        SaveStats(FieldNdx,StringGrid.Cells[i,j]);
    end;
  DisplayStatslabels;
end;

procedure TvStringData.SaveStats(const FieldIndex: integer; const Value: string);
begin
  TableData.SumTotal(FieldIndex,StrToFloatDef(Value, 0));
  TableData.SetMinimum(FieldIndex,StrToIntDef(Value, 0));
  TableData.SetMaximum(FieldIndex,StrToIntDef(Value, 0));
end;

procedure TvStringData.DisplayStatsLabels;
var
  i, FieldNdx: integer;
begin
  for i := 2 to StringGrid.ColCount - 1 do
  begin
    ClearStatsColumn(i);
    FieldNdx := ColNdx2FieldNdx(i);
    StatsGrid.Cells[i, 0] := IntToStr(TableData.GetNonZeroCount(FieldNdx));
    if TableData.GetNonZeroCount(FieldNdx) <> 0 then
      case TableData.GetFieldType(FieldNdx) of
        vString, vDescription, vCourse, vURL: ;
        vSeconds: DisplayStatsColumn(i, True)
        else      DisplayStatsColumn(i, False);
      end;
  end;
end;

procedure TvStringData.ClearStatsColumn(const Column: integer);
begin
  StatsGrid.Cells[Column, 0] := '';
  StatsGrid.Cells[Column, 1] := '';
  StatsGrid.Cells[Column, 2] := '';
  StatsGrid.Cells[Column, 3] := '';
  StatsGrid.Cells[Column, 4] := '';
end;

procedure TvStringData.DisplayStatsColumn(const Column: integer; const SecondsField: boolean);
var
  FieldNdx: integer;
begin
  FieldNdx := ColNdx2FieldNdx(Column);
  if SecondsField then begin
    StatsGrid.Cells[Column, 1] := FormatSeconds(IntToStr(TableData.GetAverage(FieldNdx)), True);
    StatsGrid.Cells[Column, 2] := FormatSeconds(FloatToStr(TableData.GetTotal(FieldNdx)), True);
    StatsGrid.Cells[Column, 3] := FormatSeconds(FloatToStr(TableData.GetMinimum(FieldNdx)), True);
    StatsGrid.Cells[Column, 4] := FormatSeconds(FloatToStr(TableData.GetMaximum(FieldNdx)), True);
  end
  else begin
    StatsGrid.Cells[Column, 1] := IntToStr(TableData.GetAverage(FieldNdx));
    StatsGrid.Cells[Column, 2] := FloatToStr(TableData.GetTotal(FieldNdx));
    StatsGrid.Cells[Column, 3] := FloatToStr(TableData.GetMinimum(FieldNdx));
    StatsGrid.Cells[Column, 4] := FloatToStr(TableData.GetMaximum(FieldNdx));
  end;
end;

procedure TvStringData.ResultsGridSelection(Sender: TObject; aCol, aRow: integer);
begin
  ShowbtnMap(arow);
  Unused(Acol);
end;

procedure TvStringData.ShowBtnMap(const Row: integer);
var
  Index: integer;
begin
  btnMap.Visible := False;
  if TableData.IsCourseTable then begin
    Index := FieldNdx2ColNdx(TableData.IndexOfField(vURL));
    if (Index >= 0) and (StringGrid.Cells[Index, row] <> '') then
      btnMap.Visible := True;
  end
  else begin
    Index := FieldNdx2ColNdx(TableData.IndexOfField(vCourse, CurrentTable));
    if (Index >= 0) and CourseData.URLExists(StringGrid.cells[Index, Row]) then
      btnMap.Visible := True;
  end;
end;

procedure TvStringData.btnMapClick(Sender: TObject);
begin
  if TableData.IsCourseTable then
    ShowMap(StringGrid.Cells[FieldNdx2ColNdx(TableData.IndexOfField(vURL)), StringGrid.Row])
  else
    ShowMap(CourseData.CourseURL(StringGrid.cells[FieldNdx2ColNdx(TableData.IndexOfField(vCourse)), StringGrid.Row]));
end;

function TvStringData.FieldNdx2ColNdx(const FieldIndex: integer): integer;
var
  i: integer = 1;
begin
  while (i < StringGrid.ColCount) and (ColNdx2FieldNdx(i) <> FieldIndex) do i := i + 1;
  if i = StringGrid.ColCount then i := -1;
  Result := i;
end;

function TvStringData.ColNdx2FieldNdx(const ColumnIndex: integer): integer;
begin
  Result := ColumnsFields[ColumnIndex];
end;

procedure TvStringData.DeleteColumn(col: integer);
begin
  StringGrid.DeleteCol(col);
  while (ColumnsFields[col] <> -1) do begin
    ColumnsFields[col] := ColumnsFields[col+1];
    col := col + 1;
  end;
end;

procedure TvStringData.InitColumnsFields;
var
  i: integer;
begin
  for i := 0 to StringGrid.ColCount - 1 do
    ColumnsFields[i] := i;
  ColumnsFields[StringGrid.ColCount] := -1;
end;

procedure TvStringData.AlignColumnsWithFields;
var
  i:integer = 1;
  Feeld: integer;
begin
  while i < StringGrid.ColCount do begin
    Feeld := TableData.IndexOfFieldGivenFieldName(StringGrid.Cells[i,0]);
    if Feeld = i then
      i := i + 1
    else
      StringGrid.MoveColRow(True, i, Feeld);
  end;
end;

end.

