unit VtableData;

{$mode objfpc} {$H+}

////////////////////////////////////////////////////////////////////////
////       Unit providing operations on the tables' metadata        ////
////////////////////////////////////////////////////////////////////////

interface

uses
  Classes, SysUtils, dialogs, StdCtrls, vINI, vSqlite3, vTypes;

type
  { TvTableData }

  TvTableData = class
  private
    TableInfo: VTableInfoArrayType;
    ti:        integer;     // table index
    tt:        string;      // Temporary table - used when getting info for other than the current table
    nt:        integer;     // Number of Tables
    INI:       TvINI;       // Pointer to the TvINI object which reads the INI file.
    vvSqlite3: TvSqlite3;   // Pointer to the TvSqlite3 object which reads data from the database
    /////// >>>>> Procedures used by Properties <<<<<< ///////
    procedure SetTableIndex(const AValue: integer);
    function GetNumberOfHiddenFields: integer;
    procedure SetTableCount(const AValue: integer);
    procedure SetTableTo(Table: string);
    function GetTableName:string;
    function IsTableCourses:boolean;
    procedure SetTemporaryTable(const AValue: string);
    /////// >>>>> Properties <<<<<< ///////
    // Index (in TableInfo) of currently selected table
    property TemporaryTable:string read tt write SetTemporaryTable;
    // Index (in TableInfo) of currently selected table
    property TableIndex:integer read ti write SetTableIndex;
    // Number of Tables
    property TableCount: integer read nt write SetTableCount;
    // Return the number of hidden fields in the current table
    property NumberOfHiddenFields: integer read GetNumberOfHiddenFields;
    /////// >>>>> Other procedures <<<<<< ///////
    // Get and save the table names, and metadata into TableInfo.
    procedure LoadTableNames;
    // Get and save the Field metadata into TableInfo
    procedure LoadFieldInfo;
    // Get and save the names of the fields from the the database
    procedure SetFieldNames;
    // Get and save the metadata for all the fields
    procedure SetFieldsMetadata;
    // Get and save the metadata for one field
    procedure SetFieldMetaData(const Tablendx,Fieldndx:integer);
    // Store the Field names in 'Fields' to the TableInfo[Tablendx] array
    procedure StoreFieldNames(const Tablendx: integer);
    // Return the index of the field given the UFfieldname
    function IndexOfField(const UFfieldName:string; const Table: string = ''):integer;
    // Change the table to tempTable so as to retrieve Table/field metadata in
    // other procedures. The calling procedure must call ChangeTable(False) at the
    // end of the procedure to set the Current Table back.
    procedure ChangeTable(const ToTemporaryTable: boolean; const TempTable: string = '');
    // If no default X or Y values, set them to fields 2 and 1 respectively
    procedure SetDefaultY(const TableNdx: integer);
    procedure SetDefaultX(const TableNdx: integer);
  public
    constructor Create(INIfile:TvINI; Sqlite3:TvSqlite3);
    destructor Destroy; override;
    /////// >>>>> Properties <<<<<< ///////
    // set/get the current table name
    property TableName: string read GetTableName write SetTableTo;
    // True if the current table is the 'courses' table
    property IsCourseTable: boolean read IsTableCourses;
    /////// >>>>> Other procedures <<<<<< ///////
    // Return the User-friendly Table Name of 'Table'
    function GetUFtableName(const Table: string = ''):string;
    // Return the Table name, given the user-friendly table name
    function GetTableName(const UFtableName: string): string;
    // The number of Fields in 'Table'
    function GetNumberOfFields(const Table: string = ''; const IncludeHidden: boolean = True):integer;
    // Get the Field Name of the ith field
    function IthFieldName(const i: integer; Table: string = ''): string;
    // Get the user-friendly Field Name of the ith field
    function ithUFfieldName(const i: integer; Table: string = ''): string;
    // Return the name of the ith Field for the table 'Table'
    function GetFieldName(const UFfieldName: string; Table: string = ''): string;
    // Return the User-friendly field name of the Ith field
    function GetUFfieldName(const FieldName: string; Table: string = ''): string;
    // Return the User-friendly name of the i-th field
    function GetithUFfieldName(const i:integer; Table: string = ''): string;
    // Return the type of field for the i-th field (eg seconds, description)
    function GetFieldType(const i:integer; Table: string = ''): VFieldType;
    function GetFieldType(const UFfieldName:string; Table: string = ''): VFieldType;
    // Return the index in fieldInfo of the currently selected table for the given Field Type
    function IndexOfField(const Field: vFieldType; Table: string = ''):integer;
    // Load the user-friendly names of the tables into a Combo Box
    procedure LoadUFtableNames(var cmb: TComboBox; const AllTables: boolean = False);
    // Load the names of the tables into a List Box
    procedure LoadTableNames(var Lst: TListBox; const AllTables: boolean = True);
    // Return all the user-friendly names of the fields of the currently selected table
    function GetUFfieldNames(const Table: string = ''; const IncludeHidden: boolean = True):VfieldArrayType;
    // Return the index in fieldInfo of the currently selected table for the Default Field to plot on the Y axis.
    function IndexOfDefYvalue(const Table: string = ''): integer;
    // Return the index in fieldInfo of the currently selected table for the Default Field to plot on the Y axis.
    function IndexOfDefXvalue(const Table: string = ''): integer;
    // Return the index of the field given the Fieldname
    function IndexOfFieldGivenFieldName(const FieldName:string; const Table: String = ''):integer;
    // set the total to zero for all fields
    procedure ZeroStatsTotals;
    // Increment and get the non-zero count for the i-th field
    procedure IncrementNZC(const i:integer);
    function GetNonZeroCount(const i:integer) :integer;
    // Set and get the minimum value for the i-th field to amt
    procedure SetMinimum(const i:integer; const amt: real);
    function GetMinimum(const i:integer):real;
    // Set and get the maxium value for the i-th field to amt
    procedure SetMaximum(const i:integer; const amt: real);
    function GetMaximum(const i:integer):real;
    // Sum and get the amt to the Total of the i-th Field
    procedure SumTotal(const i:integer; const amt:real);
    function GetTotal(const i:integer):real;
    // Get the average for the i-th field
    function GetAverage(const i:integer):integer;
    // Return the 'hidden' value for the i-th field of table 'Table'
    function GetFieldHidden(const FIeldNdx: integer; const Table: string = ''): boolean;
    // Return the 'graphed' value for the i-th field of table 'Table'
    function GetGraphed(const Table: string = ''): boolean;
    // Return a complete copy of the TableInfo data (used in editing the table structures)
    function AllTableInfo: vTableInfoArrayType;
    // Add a list of User-friendly field names to a ComboBox
    procedure VisibleUFfieldNames(cmb: TcomboBox; const Table: string = '';
      const IncludeDateCourseFields: boolean = False);
    // Return the first table in TableInfo
    function FirstTable: string;
    // Return the default values for the Table 'table'
    function GetDefaultValues(const Table: string = ''): vFieldArrayType;
    // Add the Field names to the Fields array
    procedure AddFieldNames(var fields: vFieldArrayType; const Table: string = '');
end;

implementation

uses StrUtils;

{ TvTableData }

function TvTableData.GetNumberOfHiddenFields: integer;
var
  i:  integer;
  HF: integer = 0;
begin
  for i := 1 to TableInfo[TableIndex].NF -1 do
    if TableInfo[TableIndex].FieldInfo[i].Hidden then
      HF := HF + 1;
  Result := HF;
end;

procedure TvTableData.SetTemporaryTable(const AValue: string);
begin
  tt:=AValue;
end;

procedure TvTableData.SetTableIndex(const AValue: integer);
begin
  ti:=AValue;
end;

procedure TvTableData.SetTableCount(const AValue: integer);
begin
  nt:=AValue;
end;

// Get the (User-friendly) Tablenames from the vfitness.ini and store in TableInfo
procedure TvTableData.LoadTableNames;
var
  i, j : integer;
  TableNames, UFtableNames, TableOrder: TStringList;
begin
  TableIndex      :=  0;
  TableCount      :=  0;
  TableNames      := TStringList.Create;
  UFtableNames    := TStringList.Create;
  TableOrder      := TStringList.Create;
  TableNames      := INI.TableNames;
  UFtableNames    := INI.UFtableNames;
  TableOrder      := INI.TableOrder;
  for i := 0 to TableOrder.Count - 1 do
  begin
    for j := 0 to TableNames.Count - 1 do
      if TableNames[j] = TableOrder[i] then break;
    TableInfo[i].UFN        := UFtableNames[j];
    TableInfo[i].TN         := TableNames[j];
    TableInfo[i].Hidden     := INI.IsTableHidden(TableInfo[i].TN);
    TableInfo[i].NotGraphed := not INI.IsTableGraphed(TableInfo[i].TN);
    TableInfo[i].Count      := 0;
    TableInfo[i].NF         := 0;
    TableCount              := i + 1;
    if i = MaxTables then break;
  end;
  TemporaryTable := '';
  TableNames.Destroy;
  UFtableNames.Destroy;
  TableOrder.Destroy;
end;

procedure TvTableData.LoadFieldInfo;
begin
  SetFieldNames;
  SetFieldsMetadata;
end;

constructor TvTableData.Create(INIfile:TvINI; Sqlite3:TvSqlite3);
begin
  INI             := INIfile;
  vvSqlite3       := Sqlite3;
  LoadTableNames;
  LoadFieldInfo;
end;

destructor TvTableData.Destroy;
begin
  inherited Destroy;
end;

function TvTableData.IthFieldName(const i: integer; Table: string = ''): string;
begin
  ChangeTable(True, Table);
  Result := TableInfo[TableIndex].FieldInfo[i].FN;
  ChangeTable(False);
end;

function TvTableData.ithUFfieldName(const i: integer; Table: string): string;
begin
  ChangeTable(True, Table);
  Result := TableInfo[TableIndex].FieldInfo[i].UFN;
  ChangeTable(False);
end;

function TvTableData.GetFieldName(const UFfieldName: string; Table: string): string;
begin
  ChangeTable(True, Table);
  Result := ithFieldName(IndexOfField(UFfieldName));
  ChangeTable(False);
end;

function TvTableData.GetUFfieldName(const FieldName: string; Table: string): string;
begin
  ChangeTable(True, Table);
  Result := ithUFfieldName(IndexOfFieldGivenFieldName(FieldName));
  ChangeTable(False);
end;

procedure TvTableData.SetTableTo(Table: string);
var
  i: integer = 0;
begin
  while (i < TableCount) and (CompareText(Table,TableInfo[i].TN) <> 0) do
    i := i + 1;
  TableIndex := i;
end;

function TvTableData.IsTableCourses: boolean;
begin
  Result := CompareText(TableInfo[TableIndex].TN, 'Courses') = 0;
end;

function TvTableData.IndexOfDefYvalue(const Table: string = ''): integer;
var
  i: integer = 0;
begin
  ChangeTable(True, Table);
  while (i < GetNumberOfFields) and not TableInfo[TableIndex].FieldInfo[i].Y do
    i := i + 1;
  if i = GetNumberOfFields then
    Result := -1
  else
    Result := i;
  ChangeTable(False);
end;

function TvTableData.IndexOfDefXvalue(const Table: string = ''): integer;
var
  i: integer = 0;
begin
  ChangeTable(True, Table);
  while (i < GetNumberOfFields) and not TableInfo[TableIndex].FieldInfo[i].X do
    i := i + 1;
  if i = GetNumberOfFields then
    Result := -1
  else
    Result := i;
  ChangeTable(False);
end;

function TvTableData.GetUFtableName(const Table: string = ''): string;
begin
  ChangeTable(True, Table);
  Result := TableInfo[TableIndex].UFN;
  ChangeTable(False);
end;

function TvTableData.GetFieldHidden(const FIeldNdx: integer; const Table: string): boolean;
begin
  ChangeTable(True, Table);
  Result := TableInfo[TableIndex].FieldInfo[FieldNdx].Hidden;
  ChangeTable(False);
end;

function TvTableData.GetGraphed(const Table: string): boolean;
begin
  ChangeTable(True, Table);
  Result := not TableInfo[TableIndex].NotGraphed;
  ChangeTable(False);
end;

function TvTableData.AllTableInfo: vTableInfoArrayType;
begin
  Result := TableInfo;
end;

procedure TvTableData.VisibleUFfieldNames(cmb: TcomboBox; const Table: string; const IncludeDateCourseFields: boolean = False);
var
  i: integer;
begin
  ChangeTable(True, Table);
  cmb.Clear;
  for i := 1 to TableInfo[TableIndex].NF -1 do begin
    if (IncludeDateCourseFields or not (GetFieldType(i) in [vDate, vCourse])) and not GetFieldHidden(i) then
      cmb.Items.Add(ithUFfieldName(i));
  end;
  cmb.ItemIndex := 0;
  ChangeTable(False);
end;

function TvTableData.FirstTable: string;
begin
  Result := TableInfo[0].TN;
end;

function TvTableData.GetDefaultValues(const Table: string = ''): vFieldArrayType;
var
  Fields: vFieldArrayType;
  i: integer;
begin
  ChangeTable(True, Table);
  for i := 0 to GetNumberOfFields do
    Fields[i].Value := TableInfo[TableIndex].FieldInfo[i].DefaultValue;
  Result := Fields;
  ChangeTable(False);
end;

procedure TvTableData.AddFieldNames(var fields: vFieldArrayType; const Table: string = '');
var
  i: integer = 0;
begin
  ChangeTable(True,Table);
  while i < TableInfo[TableIndex].NF do begin
    Fields[i].Name := ithFieldName(i, Table);
    i := i + 1;
  end;
  ChangeTable(False);
end;

function TvTableData.IndexOfField(const Field: vFieldType; Table: string = ''): integer;
var
  i: integer = 0;
begin
  ChangeTable(True, Table);
  while (i < GetNumberOfFields) and (GetFieldType(i, Table) <> Field) do
    i := i + 1;
  if i = GetNumberOfFields then
    Result := -1
  else
    Result := i;
  ChangeTable(False);
end;

function TvTableData.IndexOfField(const UFfieldName: string; const Table: string = ''): integer;
var
  i:integer = 0;
begin
  ChangeTable(True, Table);
  while (i < GetNumberOfFields) and (CompareText(TableInfo[TableIndex].FieldInfo[i].UFN, UFfieldName) <> 0) do
    i := i + 1;
  if i = GetNumberOfFields then
    Result := -1
  else
    Result := i;
  ChangeTable(False);
end;

function TvTableData.GetFieldType(const i:integer; Table: string = ''): VFieldType;
begin
  ChangeTable(True, Table);
  Result := TableInfo[TableIndex].FieldInfo[i].FT;
  ChangeTable(False);
end;

function TvTableData.GetFieldType(const UFfieldName: string; Table: string): VFieldType;
begin
  ChangeTable(True, Table);
  Result := GetFieldType(IndexOfField(UFfieldName));
  ChangeTable(False);
end;

function TvTableData.GetNumberOfFields(const Table: string = ''; const IncludeHidden: boolean = True):integer;
begin
  ChangeTable(True,Table);
  if IncludeHidden then
    Result := TableInfo[TableIndex].NF
  else
    Result := TableInfo[TableIndex].NF - NumberOfHiddenFields;
  ChangeTable(False);
end;

function TvTableData.GetUFfieldNames(const Table: string = ''; const IncludeHidden: boolean = True): VfieldArrayType;
var
  Fields: VfieldArrayType;
  i: integer;
begin
  ChangeTable(True, Table);
  for i := 0 to TableInfo[TableIndex].NF - 1 do
    if IncludeHidden or not TableInfo[TableIndex].FieldInfo[i].Hidden then
      Fields[i].Value:=TableInfo[TableIndex].FieldInfo[i].UFN;
  Result := Fields;
  ChangeTable(False);
end;

function TvTableData.GetTableName(const UFtableName: string): string;
var
  i: integer = 0;
begin
  while (i < TableCount) and (CompareText(TableInfo[i].UFN,UFtableName) <> 0) do
    i := i + 1;
  Result := ifthen(i < TableCount, TableInfo[i].TN, 'NoSuchTable');
end;

function TvTableData.GetTableName: string;
begin
  Result := TableInfo[TableIndex].TN;
end;

function TvTableData.GetithUFfieldName(const i:integer; Table: string = ''): string;
begin
  ChangeTable(True, Table);
  Result := TableInfo[TableIndex].FieldInfo[i].UFN;
  ChangeTable(False);
end;

procedure TvTableData.IncrementNZC(const i: integer);
begin
  TableInfo[TableIndex].FieldInfo[i].NZC := TableInfo[TableIndex].FieldInfo[i].NZC+1;
end;

procedure TvTableData.LoadUFtableNames(var cmb: TComboBox; const AllTables: boolean = False);
var
  i:integer;
begin
  cmb.Clear;
  for i := 0 to TableCount - 1 do
    if AllTables or not TableInfo[i].Hidden then
      cmb.Items.Add(TableInfo[i].UFN);
  TableIndex := 0;
  cmb.ItemIndex := TableIndex;
end;

procedure TvTableData.LoadTableNames(var Lst: TListBox; const AllTables: boolean = True);
var
  i:integer;
begin
  Lst.Clear;
  for i := 0 to TableCount - 1 do
    if AllTables or not TableInfo[i].Hidden then
      Lst.Items.Add(TableInfo[i].TN);
  TableIndex := 0;
  Lst.ItemIndex := TableIndex;
end;

procedure TvTableData.ZeroStatsTotals;
var
  i:integer;
begin
  for i := 0 to TableInfo[TableIndex].NF - 1 do
  begin
    TableInfo[TableIndex].FieldInfo[i].NZC := 0;
    TableInfo[TableIndex].FieldInfo[i].Tot := 0;
    TableInfo[TableIndex].FieldInfo[i].Min := 0;
    TableInfo[TableIndex].FieldInfo[i].Max := 0;
  end;
end;

procedure TvTableData.SetMinimum(const i: integer; const amt: real);
begin
  if TableInfo[TableIndex].FieldInfo[i].Min = 0 then
    TableInfo[TableIndex].FieldInfo[i].Min := amt
  else if (TableInfo[TableIndex].FieldInfo[i].Min > amt) and (amt <> 0) then
    TableInfo[TableIndex].FieldInfo[i].Min := amt;
end;

procedure TvTableData.SetMaximum(const i: integer; const amt: real);
begin
  if TableInfo[TableIndex].FieldInfo[i].Max < amt then
    TableInfo[TableIndex].FieldInfo[i].Max := amt;
end;

procedure TvTableData.SumTotal(const i: integer; const amt: real);
begin
  TableInfo[TableIndex].FieldInfo[i].Tot := TableInfo[TableIndex].FieldInfo[i].Tot + amt;
end;

procedure TvTableData.SetFieldNames;
var
  i: integer;
begin
  for i := 0 to TableCount -1 do begin
    vvSqlite3.FieldNames(TableInfo[i].NF,TableInfo[i].TN);
    StoreFieldNames(i);
  end;
end;

procedure TvTableData.SetFieldsMetadata;
var
  Tablendx, Fieldndx: integer;
begin
  for Tablendx := 0 to TableCount - 1 do begin
    TableInfo[Tablendx].FieldInfo[2].Y := True;
    TableInfo[Tablendx].FieldInfo[1].X := True;
    for Fieldndx := 0 to TableInfo[Tablendx].NF -1 do
      SetFieldMetaData(Tablendx,Fieldndx);
    SetDefaultY(Tablendx);
    SetDefaultX(Tablendx);
  end;
end;

procedure TvTableData.SetFieldMetaData(const Tablendx, Fieldndx: integer);
var
  Afield: VfieldInfoType;
begin
  Afield := INI.GetFieldInfo(TableInfo[Tablendx].TN,TableInfo[Tablendx].FieldInfo[Fieldndx].FN);
  TableInfo[Tablendx].FieldInfo[Fieldndx].DefaultValue :=Afield.DefaultValue;
  TableInfo[Tablendx].FieldInfo[Fieldndx].FT           :=Afield.FT;
  TableInfo[Tablendx].FieldInfo[Fieldndx].Max          :=Afield.Max;
  TableInfo[Tablendx].FieldInfo[Fieldndx].Min          :=Afield.Min;
  TableInfo[Tablendx].FieldInfo[Fieldndx].NZC          :=Afield.NZC;
  TableInfo[Tablendx].FieldInfo[Fieldndx].Tot          :=Afield.Tot;
  TableInfo[Tablendx].FieldInfo[Fieldndx].UFN          :=Afield.UFN;
  TableInfo[Tablendx].FieldInfo[Fieldndx].Hidden       :=Afield.Hidden;
 // If Afield.Y {and (Fieldndx <> 2)} then begin
    TableInfo[Tablendx].FieldInfo[Fieldndx].Y          :=Afield.Y;
//    TableInfo[Tablendx].FieldInfo[2].Y                 :=False;
//  end;
//  If Afield.X {and (Fieldndx <> 1)} then begin
    TableInfo[Tablendx].FieldInfo[Fieldndx].X          :=Afield.X;
//    TableInfo[Tablendx].FieldInfo[1].X                 :=False;
//  end;
end;

procedure TvTableData.StoreFieldNames(const Tablendx: integer);
var
  i:integer;
  FieldOrder: TstringList;
begin
  FieldOrder := TstringList.Create;
  FieldOrder := INI.GetFieldOrder(TableInfo[Tablendx].TN);
  for i := 0 to TableInfo[Tablendx].NF -1 do begin
    if FieldOrder[i] = '' then break;
    TableInfo[Tablendx].FieldInfo[i].FN := FieldOrder[i];
  end;
  FieldOrder.Destroy;
end;

function TvTableData.IndexOfFieldGivenFieldName(const FieldName: string; const Table: String = ''): integer;
var
  i:integer;
begin
  ChangeTable(True, Table);
  Result := -1;
  for i := 0 to GetNumberOfFields -1 do begin
    if CompareText(TableInfo[TableIndex].FieldInfo[i].FN, FieldName) = 0 then begin
      Result := i;
      exit;
    end;
  end;
  ChangeTable(False);
end;

procedure TvTableData.ChangeTable(const ToTemporaryTable: boolean; const TempTable: string = '');
begin
  if toTemporaryTable then begin
    if (TempTable <> '') and (TemporaryTable = '') then begin
      TemporaryTable := TableName;
      TableName      := TempTable;
    end
  end
  else if TemporaryTable <> '' then begin
    TableName := TemporaryTable;
    TemporaryTable := '';
  end;
end;

procedure TvTableData.SetDefaultY(const TableNdx: integer);
var
  i: integer;
begin
  for i := 0 to MaxFields do
    if TableInfo[Tablendx].FieldInfo[i].Y then
      exit;
  TableInfo[Tablendx].FieldInfo[2].Y:=True;
end;

procedure TvTableData.SetDefaultX(const TableNdx: integer);
var
  i: integer;
begin
  for i := 0 to MaxFields do
    if TableInfo[Tablendx].FieldInfo[i].X then
      exit;
  TableInfo[Tablendx].FieldInfo[1].X:=True;
end;

function TvTableData.GetNonZeroCount(const i: integer): integer;
begin
  Result := TableInfo[TableIndex].FieldInfo[i].NZC;
end;

function TvTableData.GetMinimum(const i: integer): real;
begin
  Result := TableInfo[TableIndex].FieldInfo[i].Min;
end;

function TvTableData.GetMaximum(const i: integer): real;
begin
  Result := TableInfo[TableIndex].FieldInfo[i].Max;
end;

function TvTableData.GetTotal(const i: integer): real;
begin
  Result := TableInfo[TableIndex].FieldInfo[i].Tot;
end;

function TvTableData.GetAverage(const i: integer): integer;
begin
  Result := round(GetTotal(i) / GetNonZeroCount(i));
end;

end.

